import logging
from collections import namedtuple

import django.db
from django import VERSION
from django.apps import apps
from django.db import models, migrations
from django.db.migrations.migration import Migration
from django.db.migrations.state import ProjectState
from django.db.models import UniqueConstraint
from django.db.utils import DEFAULT_DB_ALIAS, ConnectionHandler, ProgrammingError
from django.test import TestCase, TransactionTestCase
from unittest import skipIf, expectedFailure

from . import get_constraints
from ..models import (
    TestIndexesRetainedRenamed,
    Choice,
    Question,
)

connections = ConnectionHandler()

if (VERSION >= (3, 2)):
    from django.utils.connection import ConnectionProxy
    connection = ConnectionProxy(connections, DEFAULT_DB_ALIAS)
else:
    from django.db import DefaultConnectionProxy
    connection = DefaultConnectionProxy()

logger = logging.getLogger('mssql.tests')

# Result type for migration test helper
MigrationTestResult = namedtuple('MigrationTestResult', ['model', 'constraints', 'project_state'])


class TestIndexesRetained(TestCase):
    """
    Issue https://github.com/microsoft/mssql-django/issues/14
    Indexes dropped during a migration should be re-created afterwards
    assuming the field still has `db_index=True`
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        # Pre-fetch which indexes exist for the relevant test model
        # now that all the test migrations have run
        cls.constraints = get_constraints(table_name=TestIndexesRetainedRenamed._meta.db_table)
        cls.indexes = {k: v for k, v in cls.constraints.items() if v['index'] is True}

    def _assert_index_exists(self, columns):
        matching = {k: v for k, v in self.indexes.items() if set(v['columns']) == columns}
        assert len(matching) == 1, (
            "Expected 1 index for columns %s but found %d %s" % (
                columns,
                len(matching),
                ', '.join(matching.keys())
            )
        )

    def test_field_made_nullable(self):
        # case (a) of https://github.com/microsoft/mssql-django/issues/14
        self._assert_index_exists({'a'})

    def test_field_renamed(self):
        # case (b) of https://github.com/microsoft/mssql-django/issues/14
        self._assert_index_exists({'b_renamed'})

    def test_table_renamed(self):
        # case (c) of https://github.com/microsoft/mssql-django/issues/14
        self._assert_index_exists({'c'})

def _get_all_models():
    for app in apps.get_app_configs():
        app_label = app.label
        for model_name, model_class in app.models.items():
            yield model_class, model_name, app_label


class TestCorrectIndexes(TestCase):

    def test_correct_indexes_exist(self):
        """
        Check there are the correct number of indexes for each field after all migrations
        by comparing what the model says (e.g. `db_index=True` / `index_together` etc.)
        with the actual constraints found in the database.
        This acts as a general regression test for issues such as:
         - duplicate index created (e.g. https://github.com/microsoft/mssql-django/issues/77)
         - index dropped but accidentally not recreated
         - index incorrectly 'recreated' when it was never actually dropped or required at all
        Note of course that it only covers cases which exist in testapp/models.py and associated migrations
        """
        connection = django.db.connections[django.db.DEFAULT_DB_ALIAS]
        for model_cls, model_name, app_label in _get_all_models():
            logger.debug('Checking model: %s.%s', app_label, model_name)
            if not model_cls._meta.managed:
                # Models where the table is not managed by Django migrations are irrelevant
                continue
            model_constraints = get_constraints(table_name=model_cls._meta.db_table)
            # Check correct indexes are in place for all fields in model
            for field in model_cls._meta.get_fields():
                if not hasattr(field, 'column'):
                    # ignore things like reverse fields which don't have a column on this table
                    continue
                col_name = connection.introspection.identifier_converter(field.column)
                field_str = f'{app_label}.{model_name}.{field.name} ({col_name})'
                logger.debug('  > Checking field: %s', field_str)

                # Find constraints which include this column
                col_constraints = [
                    dict(name=name, **infodict) for name, infodict in model_constraints.items()
                    if col_name in infodict['columns']
                ]
                col_indexes = [c for c in col_constraints if c['index']]
                for c in col_constraints:
                    logger.debug('    > Column <%s> is involved in constraint: %s', col_name, c)

                # There should be an explicit index for each of the following cases
                expected_index_causes = []
                if field.db_index:
                    expected_index_causes.append('db_index=True')
                if VERSION < (5, 1):
                   for field_names in model_cls._meta.index_together:
                      if field.name in field_names:
                         expected_index_causes.append(f'index_together[{field_names}]')
                if field._unique and field.null:
                    # This is implemented using a (filtered) unique index (not a constraint) to get ANSI NULL behaviour
                    expected_index_causes.append('unique=True & null=True')
                for field_names in model_cls._meta.unique_together:
                    if field.name in field_names:
                        # unique_together results in an index because this backend implements it using a
                        # (filtered) unique index rather than a constraint, to get ANSI NULL behaviour
                        expected_index_causes.append(f'unique_together[{field_names}]')
                for uniq_constraint in filter(lambda c: isinstance(c, UniqueConstraint), model_cls._meta.constraints):
                    if field.name in uniq_constraint.fields and uniq_constraint.condition is not None:
                        # Meta:constraints > UniqueConstraint with condition are implemented with filtered unique index
                        expected_index_causes.append(f'UniqueConstraint (with condition) in Meta: constraints')

                # Other cases like `unique=True, null=False` or `field.primary_key` do have index-like constraints
                # but in those cases the introspection returns `"index": False` so they are not in the list of
                # explicit indexes which we are checking here (`col_indexes`)

                assert len(col_indexes) == len(expected_index_causes), \
                    'Expected %s index(es) on %s but found %s.\n' \
                    'Check for behaviour changes around index drop/recreate in methods like _alter_field.\n' \
                    'Expected due to: %s\n' \
                    'Found: %s' % (
                        len(expected_index_causes),
                        field_str,
                        len(col_indexes),
                        expected_index_causes,
                        '\n'.join(str(i) for i in col_indexes),
                    )
                logger.debug('  Found %s index(es) as expected', len(col_indexes))


class TestIndexesBeingDropped(TestCase):

    def test_unique_index_dropped(self):
        """
        Issues https://github.com/microsoft/mssql-django/issues/110
        and https://github.com/microsoft/mssql-django/issues/90
        Unique indexes not being dropped when changing non-nullable
        foreign key with unique_together to nullable causing
        dependent on column error
        """
        old_field = Choice._meta.get_field('question')
        new_field = models.ForeignKey(
            Question, null=False, on_delete=models.deletion.CASCADE
        )
        new_field.set_attributes_from_name("question")
        with connection.schema_editor() as editor:
            editor.alter_field(Choice, old_field, new_field, strict=True)

        old_field = new_field
        new_field = models.ForeignKey(
            Question, null=True, on_delete=models.deletion.CASCADE
        )
        new_field.set_attributes_from_name("question")
        try:
            with connection.schema_editor() as editor:
                editor.alter_field(Choice, old_field, new_field, strict=True)
        except ProgrammingError:
            self.fail("Unique indexes not being dropped")

class TestMetaIndexesRetained(TransactionTestCase):
    """
    Regression test for indexes defined via Meta.indexes being dropped
    and not recreated after altering one of the indexed columns.

    Tests various schema operations that trigger index drop/recreate logic to ensure
    indexes are properly restored.

    Each test runs twice:
    - With migrations in split contexts (simulates separate migration files)
    - With migrations in combined context (simulates single migration file with multiple operations)
    """

    def _run_migration_test(
        self,
        operations_a: list,
        operations_b: list,
        migration_name_prefix: str,
        model_name: str,
        use_single_migration: bool,
    ) -> MigrationTestResult:
        """
        Helper to run migration tests with either combined or split schema_editor contexts.

        Args:
            operations_a: List of operations for initial setup (CreateModel + AddIndex)
            operations_b: List of operations for the alteration being tested
            migration_name_prefix: Prefix for migration names (e.g., 'test_mc_type')
            model_name: Name of the model being tested
            use_single_migration: If True, combine both operation lists into one Migration;
                               If False, create two separate Migrations

        Returns:
            MigrationTestResult: Named tuple containing (model, constraints, project_state)
        """
        # Use django.db.connections to get a fresh connection for TransactionTestCase
        conn = django.db.connections[django.db.DEFAULT_DB_ALIAS]
        suffix = '_combined' if use_single_migration else '_split'

        if use_single_migration:
            # Combined: Create ONE migration with all operations combined
            # This simulates combining operations in a single migration file
            class CombinedMigration(migrations.Migration):
                initial = True
                operations = operations_a + operations_b

            migration = CombinedMigration(name=f'{migration_name_prefix}{suffix}', app_label='testapp')

            with conn.schema_editor(atomic=True) as editor:
                project_state = migration.apply(ProjectState(), editor)
        else:
            # Split: Create TWO separate migrations, each with its own operations
            # This simulates two separate migration files where the first migration
            # is fully committed and `deferred_sql` runs before starting the second migration
            class MigrationA(migrations.Migration):
                initial = True
                operations = operations_a

            class MigrationB(migrations.Migration):
                operations = operations_b

            migration_a = MigrationA(name=f'{migration_name_prefix}{suffix}_a', app_label='testapp')
            migration_b = MigrationB(name=f'{migration_name_prefix}{suffix}_b', app_label='testapp')

            with conn.schema_editor(atomic=True) as editor:
                project_state = migration_a.apply(ProjectState(), editor)
            with conn.schema_editor(atomic=True) as editor:
                project_state = migration_b.apply(project_state, editor)

        # Get the model and constraints for assertions
        model = project_state.apps.get_model('testapp', model_name)
        constraints = get_constraints(table_name=model._meta.db_table)

        return MigrationTestResult(model, constraints, project_state)

    def _assert_index_exists(self, constraints, expected_columns, error_msg):
        """
        Assert that an index with exactly the expected columns exists.

        Args:
            constraints: Dictionary of constraints from get_constraints()
            expected_columns: Set of column names that should be in the index
            error_msg: Message to display if assertion fails
        """
        found = any(
            set(info['columns']) == expected_columns and info['index']
            for info in constraints.values()
        )
        self.assertTrue(found, error_msg)

    def _get_context_description(self, use_single_migration: bool) -> str:
        return "combined single migration" if use_single_migration else "split into 2 migrations"

    def test_index_from_meta_indexes_retained_after_type_change(self):
        """
        Test that indexes defined in Meta.indexes are retained when altering field type (max_length change).
        This exercises the type change code path in _alter_field.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxType{suffix}'

                operations_a = [
                    migrations.CreateModel(
                        name=model_name,
                        fields=[
                            ('id', models.AutoField(primary_key=True)),
                            ('a', models.CharField(max_length=20)),
                            ('b', models.CharField(max_length=20)),
                        ],
                    ),
                    migrations.AddIndex(
                        model_name=model_name.lower(),
                        index=models.Index(fields=['a', 'b'], name=f'idx_type{suffix}'),
                    ),
                ]

                operations_b = [
                    migrations.AlterField(
                        model_name=model_name.lower(),
                        name='a',
                        field=models.CharField(max_length=40),
                    ),
                ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_type',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after field type change "
                        f"({self._get_context_description(use_single_migration)}). Expected index to be restored after ALTER COLUMN operation."
                    ),
                )

    def test_index_from_meta_indexes_retained_after_nullability_change(self):
        """
        Test that indexes defined in Meta.indexes are retained when changing field nullability.
        This exercises the nullability change code path in _alter_field.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxNull{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_null{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='b',
                            field=models.CharField(max_length=20, null=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_null',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after nullability change "
                        f"({self._get_context_description(use_single_migration)}). Expected index to be restored after ALTER COLUMN NULL operation."
                    ),
                )

    def test_db_index_retained_after_nullability_only_change(self):
        """
        Test that db_index=True indexes are retained when ONLY nullability changes.

        This tests the case where:
        - Field has db_index=True
        - Field nullability changes (null=False → null=True)
        - Field type does NOT change (same max_length)

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestDbIndexNullChange{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20, db_index=True)),  # db_index=True, null=False
                            ],
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=20, db_index=True, null=True),  # Same type, different null
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_dbidx_null',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Verify db_index=True index was retained
                # Look for single-column index on 'a'
                db_index_indexes = [
                    info for info in result.constraints.values()
                    if info.get('index') and set(info['columns']) == {'a'}
                ]
                self.assertTrue(
                    len(db_index_indexes) > 0,
                    f"db_index=True index on 'a' was not retained after nullability-only change "
                    f"({self._get_context_description(use_single_migration)}). "
                    f"Expected index from db_index=True to be restored after changing null=False to null=True."
                )

    def test_db_index_retained_after_nullability_change_to_not_null(self):
        """
        Test that db_index=True indexes are retained when changing from null=True to null=False.

        This is the reverse direction of test_db_index_retained_after_nullability_only_change
        and exercises the four-way default alteration path in _alter_field (requires a default value).

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestDbIndexNotNull{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20, db_index=True, null=True)),  # db_index=True, null=True
                            ],
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=20, db_index=True, null=False, default=''),  # null=False requires default
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_dbidx_notnull',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Verify db_index=True index was retained
                db_index_indexes = [
                    info for info in result.constraints.values()
                    if info.get('index') and set(info['columns']) == {'a'}
                ]
                self.assertTrue(
                    len(db_index_indexes) > 0,
                    f"db_index=True index on 'a' was not retained after nullability change from NULL to NOT NULL "
                    f"({self._get_context_description(use_single_migration)}). "
                    f"Expected index from db_index=True to be restored after four-way default alteration."
                )

    def test_index_from_meta_indexes_retained_after_field_rename(self):
        """
        Test that indexes defined in Meta.indexes are retained and updated when renaming a field.
        The index should exist on the renamed column.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxRename{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_rename{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.RenameField(
                            model_name=model_name.lower(),
                            old_name='a',
                            new_name='a_renamed',
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_rename',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a_renamed', 'b'},
                    error_msg=(
                        f"Index on ('a_renamed', 'b') from Meta.indexes was not found after field rename "
                        f"({self._get_context_description(use_single_migration)}). Expected index to be updated to reflect the renamed column."
                    ),
                )

    @expectedFailure
    def test_index_from_meta_indexes_retained_after_rename_and_type_change(self):
        """
        Test that indexes from Meta.indexes are retained when a field is renamed
        AND has its type changed in the same migration.

        This tests a known bug: the TYPE CHANGE PATH drops indexes, but the
        RESTORE PHASE is skipped when column is renamed (old_field.column != new_field.column).

        Additionally, _delete_indexes() fails with FieldDoesNotExist because it tries to
        look up the index field by the old field name, but RenameField has already updated
        the model state, so the old field name no longer exists.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxRenameType{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_rename_type{suffix}'),
                        ),
                    ]

                operations_b = [
                        # Rename field 'a' to 'a_renamed'
                        migrations.RenameField(
                            model_name=model_name.lower(),
                            old_name='a',
                            new_name='a_renamed',
                        ),
                        # Also change its type (max_length 20 -> 40)
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a_renamed',
                            field=models.CharField(max_length=40),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_rename_type',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a_renamed', 'b'},
                    error_msg=(
                        f"Index on ('a_renamed', 'b') from Meta.indexes was not found after field rename + type change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be retained when both rename and type change occur."
                    ),
                )

    def test_db_index_retained_after_rename_and_type_change(self):
        """
        Test that db_index=True indexes are retained when a field's db_column is changed
        AND has its type changed in the same AlterField operation.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestDbIdxRenameType{suffix}'

                operations_a = [
                    migrations.CreateModel(
                        name=model_name,
                        fields=[
                            ('id', models.AutoField(primary_key=True)),
                            ('a', models.CharField(max_length=20, db_index=True, db_column='col_a')),
                            ('b', models.CharField(max_length=20)),
                        ],
                    ),
                ]

                operations_b = [
                    # Change db_column AND type in single AlterField
                    migrations.AlterField(
                        model_name=model_name.lower(),
                        name='a',
                        field=models.CharField(max_length=40, db_index=True, db_column='col_a_renamed'),
                    ),
                ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_dbidx_rename_type',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'col_a_renamed'},
                    error_msg=(
                        f"db_index=True index on 'col_a_renamed' was not found after column rename + type change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be retained when both column rename and type change occur in same AlterField."
                    ),
                )

    def test_unique_retained_after_rename_and_type_change(self):
        """
        Test that unique=True constraints are retained when a field's db_column is changed
        AND has its type changed in the same AlterField operation.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestUniqueRenameType{suffix}'

                operations_a = [
                    migrations.CreateModel(
                        name=model_name,
                        fields=[
                            ('id', models.AutoField(primary_key=True)),
                            ('a', models.CharField(max_length=20, unique=True, db_column='col_a')),
                            ('b', models.CharField(max_length=20)),
                        ],
                    ),
                ]

                operations_b = [
                    # Change db_column AND type in single AlterField
                    migrations.AlterField(
                        model_name=model_name.lower(),
                        name='a',
                        field=models.CharField(max_length=40, unique=True, db_column='col_a_renamed'),
                    ),
                ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_uniq_rename_type',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Check for unique constraint on the renamed column
                unique_constraints = [
                    info for info in result.constraints.values()
                    if info.get('unique') and set(info['columns']) == {'col_a_renamed'}
                ]
                self.assertTrue(
                    len(unique_constraints) > 0,
                    f"unique=True constraint on 'col_a_renamed' was not found after column rename + type change "
                    f"({self._get_context_description(use_single_migration)}). "
                    f"Expected unique constraint to be retained."
                )

    @expectedFailure
    @skipIf(VERSION >= (5, 1), "unique_together is deprecated in Django 5.1+")
    def test_unique_together_retained_after_rename_and_type_change(self):
        """
        Test that unique_together constraints are retained when a field's db_column is changed
        AND has its type changed in the same AlterField operation.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestUniqTogetherRenameType{suffix}'

                operations_a = [
                    migrations.CreateModel(
                        name=model_name,
                        fields=[
                            ('id', models.AutoField(primary_key=True)),
                            ('a', models.CharField(max_length=20, db_column='col_a')),
                            ('b', models.CharField(max_length=20)),
                        ],
                        options={
                            'unique_together': {('a', 'b')},
                        },
                    ),
                ]

                operations_b = [
                    # Change db_column AND type in single AlterField
                    migrations.AlterField(
                        model_name=model_name.lower(),
                        name='a',
                        field=models.CharField(max_length=40, db_column='col_a_renamed'),
                    ),
                ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_uniqtog_rename_type',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Check for unique_together constraint on ('col_a_renamed', 'b')
                unique_constraints = [
                    info for info in result.constraints.values()
                    if info.get('unique') and set(info['columns']) == {'col_a_renamed', 'b'}
                ]
                self.assertTrue(
                    len(unique_constraints) > 0,
                    f"unique_together constraint on ('col_a_renamed', 'b') was not found after column rename + type change "
                    f"({self._get_context_description(use_single_migration)}). "
                    f"Expected unique_together to be retained."
                )

    @expectedFailure
    def test_index_from_meta_indexes_retained_after_rename_and_nullability_change(self):
        """
        Test that indexes from Meta.indexes are retained when a field is renamed
        AND has its nullability changed in the same migration.

        This tests a known bug: the NULLABILITY CHANGE PATH drops indexes, but the
        RESTORE PHASE is skipped when column is renamed (old_field.column != new_field.column).

        Additionally, _delete_indexes() fails with FieldDoesNotExist because it tries to
        look up the index field by the old field name, but RenameField has already updated
        the model state, so the old field name no longer exists.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxRenameNull{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_rename_null{suffix}'),
                        ),
                    ]

                operations_b = [
                        # Rename field 'a' to 'a_renamed'
                        migrations.RenameField(
                            model_name=model_name.lower(),
                            old_name='a',
                            new_name='a_renamed',
                        ),
                        # Also change its nullability
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a_renamed',
                            field=models.CharField(max_length=20, null=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_rename_null',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a_renamed', 'b'},
                    error_msg=(
                        f"Index on ('a_renamed', 'b') from Meta.indexes was not found after field rename + nullability change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be retained when both rename and nullability change occur."
                    ),
                )

    def test_index_from_meta_indexes_retained_after_altering_both_fields(self):
        """
        Test that indexes defined in Meta.indexes are retained when altering multiple fields in the index.
        This ensures the index is properly restored even when both participating columns are altered.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxBoth{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_both{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40),
                        ),
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='b',
                            field=models.CharField(max_length=30),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_both',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after altering both fields "
                        f"({self._get_context_description(use_single_migration)}). Expected index to be restored after multiple ALTER COLUMN operations."
                    ),
                )

    def test_three_column_index_retained_after_field_alteration(self):
        """
        Test that indexes with 3+ columns are retained when altering one of the fields.
        This ensures the fix works for indexes with more than 2 columns.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:
            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdx3Col{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                                ('c', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b', 'c'], name=f'idx_3col{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='b',
                            field=models.CharField(max_length=50),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_3col',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b', 'c'},
                    error_msg=(
                        f"Three-column index on ('a', 'b', 'c') was not recreated after field alteration "
                        f"({self._get_context_description(use_single_migration)}). Expected index to be restored after ALTER COLUMN operation on middle column."
                    ),
                )

    def test_indexes_retained_for_field_with_db_index_and_meta_indexes(self):
        """
        Test that when a field has indexes from both db_index=True and Meta.indexes, those
        indexes are both retained after altering that field.
        """
        for use_single_migration in [False, True]:
            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxDbIdx{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20, db_index=True)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_dbidx{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40, db_index=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_dbidx',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Check that _meta_indexes index was recreated
                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after field type change "
                        f"({self._get_context_description(use_single_migration)})."
                    ),
                )

                # Check that index from db_index=True was also recreated
                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a'},
                    error_msg=(
                        "Index on 'a' from db_index=True was not recreated "
                        f"after field type change ({self._get_context_description(use_single_migration)})."
                    ),
                )

    def test_index_from_meta_indexes_retained_after_type_and_nullability_change(self):
        """
        Test that indexes defined in Meta.indexes are retained when BOTH type and nullability change simultaneously.
        This exercises both code paths in _alter_field (type change AND nullability change).
        The index should only be dropped once and recreated once (tests deduplication logic).
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxTypeNull{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_typenull{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40, null=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_typenull',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after simultaneous type and nullability change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be restored after ALTER COLUMN operation changing both max_length and nullability."
                    ),
                )

    def test_indexes_from_meta_indexes_retained_with_unique_together(self):
        """
        Test that indexes defined in Meta.indexes coexist properly with unique_together constraints.
        Tests the case where a model has overlapping columns participating in both unique_together and
        indexes defined in Meta.indexes. The index defined in Meta.indexes should be retained after field alteration.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxUniqTogether{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                                ('c', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AlterUniqueTogether(
                            name=model_name.lower(),
                            unique_together={('a', 'b')},
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'c'], name=f'idx_uniqtog{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_uniqtog',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Check that the index (a, c) from Meta.indexes was recreated
                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'c'},
                    error_msg=(
                        f"Index on ('a', 'c') from Meta.indexes was not recreated after field alteration "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to coexist with unique_together constraint on ('a', 'b')."
                    ),
                )

                # Also verify that unique_together constraint still exists
                unique_constraints = [
                    info for info in result.constraints.values()
                    if info.get('unique') and set(info['columns']) == {'a', 'b'}
                ]
                self.assertTrue(
                    len(unique_constraints) > 0,
                    f"unique_together constraint on ('a', 'b') was lost "
                    f"({self._get_context_description(use_single_migration)})."
                )

    def test_index_from_meta_indexes_retained_after_fk_alteration(self):
        """
        Test that indexes defined in Meta.indexes containing ForeignKey fields are retained after FK alteration.
        ForeignKey handling in _alter_field is complex, and this ensures that indexes defined in Meta.indexes
        involving FK fields are properly restored.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                ref_model_name = f'TestMetaIdxFKRef{suffix}'
                model_name = f'TestMetaIdxFK{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=ref_model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('name', models.CharField(max_length=50)),
                            ],
                        ),
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('fk_field', models.ForeignKey(
                                    to=f'testapp.{ref_model_name}',
                                    on_delete=models.CASCADE,
                                )),
                                ('other_field', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['fk_field', 'other_field'], name=f'idx_fk{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='fk_field',
                            field=models.ForeignKey(
                                to=f'testapp.{ref_model_name}',
                                on_delete=models.SET_NULL,
                                null=True,
                            ),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_fk',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'fk_field_id', 'other_field'},
                    error_msg=(
                        f"Index on ('fk_field', 'other_field') from Meta.indexes was not recreated after FK alteration "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be restored after changing FK from CASCADE to SET_NULL with null=True."
                    ),
                )

    def test_multiple_index_from_meta_indexes_retained(self):
        """
        Test that ALL indexes defined in Meta.indexes are retained when a field participates in multiple indexes.
        A field can be part of multiple different indexes defined in Meta.indexes, and all should be restored
        after altering that field.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaMulti{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                                ('c', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_multi_ab{suffix}'),
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'c'], name=f'idx_multi_ac{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_multi',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Check that both indexes defined in Meta.indexes were recreated
                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after field alteration "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected BOTH indexes containing field 'a' to be restored."
                    ),
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'c'},
                    error_msg=(
                        f"Index on ('a', 'c') from Meta.indexes was not recreated after field alteration "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected BOTH indexes containing field 'a' to be restored."
                    ),
                )

    def test_index_from_meta_indexes_retained_after_nullability_change_to_not_null(self):
        """
        Test that indexes defined in Meta.indexes are retained when changing field from NULL to NOT NULL.
        This is the reverse direction of the existing nullability test and exercises the
        four-way default alteration path in _alter_field (requires a default value).
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxNotNull{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20, null=True)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_notnull{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=20, null=False, default=''),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_notnull',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after nullability change from NULL to NOT NULL "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be restored after ALTER COLUMN operation with default value handling."
                    ),
                )

    def test_autofield_type_change_preserves_indexes(self):
        """
        Test that indexes defined in Meta.indexes are retained when changing AutoField to BigAutoField.
        This exercises the AutoField/BigAutoField restoration path in _alter_field
        which restores ALL indexes on ALL fields, not just the altered field.
        Runs with both split and combined migrations.
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxAutoField{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['a', 'b'], name=f'idx_auto{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='id',
                            field=models.BigAutoField(primary_key=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_auto',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"Index on ('a', 'b') from Meta.indexes was not recreated after AutoField to BigAutoField change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be restored via AutoField/BigAutoField special restoration path."
                    ),
                )

    def test_autofield_to_bigautofield_with_other_db_index_field_split(self):
        """
        Test that changing AutoField to BigAutoField preserves db_index=True indexes
        on other fields when operations are in separate migrations.

        This test verifies the split migration case works correctly - the index is
        created and committed in the first migration before the AutoField alteration
        runs in the second migration.
        """
        model_name = 'TestAutoDbIndex_split'

        operations_a = [
            migrations.CreateModel(
                name=model_name,
                fields=[
                    ('id', models.AutoField(primary_key=True)),
                    ('name', models.CharField(max_length=100, db_index=True)),
                    ('other', models.CharField(max_length=100)),
                ],
            ),
        ]

        operations_b = [
            migrations.AlterField(
                model_name=model_name.lower(),
                name='id',
                field=models.BigAutoField(primary_key=True),
            ),
        ]

        result = self._run_migration_test(
            operations_a=operations_a,
            operations_b=operations_b,
            migration_name_prefix='test_auto_dbindex',
            model_name=model_name,
            use_single_migration=False,
        )

        # Verify db_index=True index on 'name' was retained
        self._assert_index_exists(
            result.constraints,
            expected_columns={'name'},
            error_msg=(
                "db_index=True index on 'name' was not retained after AutoField to BigAutoField change "
                "(split into 2 migrations). Expected index to be preserved."
            ),
        )

    def test_autofield_to_bigautofield_with_other_db_index_field_combined(self):
        """
        Test that changing AutoField to BigAutoField preserves db_index=True indexes
        on other fields when operations are in a single (combined) migration.

        This tests that the deduplication logic in _alter_field works correctly:
        when CreateModel queues a db_index in deferred_sql and the AutoField
        restoration code runs, it should skip creating duplicate indexes.
        """
        model_name = 'TestAutoDbIndex_combined'

        operations_a = [
            migrations.CreateModel(
                name=model_name,
                fields=[
                    ('id', models.AutoField(primary_key=True)),
                    ('name', models.CharField(max_length=100, db_index=True)),
                    ('other', models.CharField(max_length=100)),
                ],
            ),
        ]

        operations_b = [
            migrations.AlterField(
                model_name=model_name.lower(),
                name='id',
                field=models.BigAutoField(primary_key=True),
            ),
        ]

        result = self._run_migration_test(
            operations_a=operations_a,
            operations_b=operations_b,
            migration_name_prefix='test_auto_dbindex',
            model_name=model_name,
            use_single_migration=True,
        )

        # Verify db_index=True index on 'name' was retained
        self._assert_index_exists(
            result.constraints,
            expected_columns={'name'},
            error_msg=(
                "db_index=True index on 'name' was not retained after AutoField to BigAutoField change "
                "(combined into 1 migration). Expected index to be preserved."
            ),
        )

    @skipIf(VERSION >= (5, 1), "index_together is removed in Django 5.1+")
    def test_index_together_retained_after_autofield_change(self):
        """
        Test that index_together indexes are retained when changing AutoField to BigAutoField.

        This tests the index_together restoration path in _alter_field for AutoField changes.
        Since AutoField changes drop ALL indexes on the table, the restoration code must
        also restore ALL index_together indexes, not just those involving the altered field.

        Note: index_together is deprecated in Django 4.2 and removed in Django 5.1+.
        This test only runs on Django < 5.1.
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestIdxTogetherAuto{suffix}'

                # Create model with index_together using the deprecated Meta option
                # We need to use a raw SQL approach or create the model dynamically
                # since Django's migration system handles index_together
                operations_a = [
                    migrations.CreateModel(
                        name=model_name,
                        fields=[
                            ('id', models.AutoField(primary_key=True)),
                            ('a', models.CharField(max_length=20)),
                            ('b', models.CharField(max_length=20)),
                        ],
                        options={
                            'index_together': {('a', 'b')},
                        },
                    ),
                ]

                operations_b = [
                    migrations.AlterField(
                        model_name=model_name.lower(),
                        name='id',
                        field=models.BigAutoField(primary_key=True),
                    ),
                ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_idx_together_auto',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"index_together index on ('a', 'b') was not recreated after AutoField to BigAutoField change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index to be restored via AutoField restoration path."
                    ),
                )

    def test_pk_type_change_preserves_indexes(self):
        """
        Test that indexes defined in Meta.indexes are retained when changing primary key type.
        This tests the primary key restoration path alongside the restoration of indexes from Meta.indexes.
        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestMetaIdxPK{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AddIndex(
                            model_name=model_name.lower(),
                            index=models.Index(fields=['id', 'a'], name=f'idx_pk{suffix}'),
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='id',
                            field=models.BigAutoField(primary_key=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_mc_pk',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Verify primary key still exists
                pk_constraints = [
                    info for info in result.constraints.values()
                    if info.get('primary_key')
                ]
                self.assertTrue(
                    len(pk_constraints) > 0,
                    f"Primary key was not restored ({self._get_context_description(use_single_migration)})."
                )

                # Verify index from Meta.indexes including PK column was restored
                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'id', 'a'},
                    error_msg=(
                        f"Index on ('id', 'a') from Meta.indexes was not recreated after PK type change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index containing PK column to be restored."
                    ),
                )

    @skipIf(VERSION >= (5, 1), "index_together removed in Django 5.1")
    def test_index_together_retained_after_type_change(self):
        """
        Test that index_together indexes are retained when altering a field type.

        IMPORTANT: This test documents the known limitation that index_together is only
        restored when the field does NOT have db_index=True. If a field has both
        db_index=True AND is in index_together, only the index from db_index=True is restored
        through the standard restoration path. This is intentional behavior for the
        deprecated index_together API (removed in Django 5.1+).

        This test uses a field WITHOUT db_index=True to verify the index_together
        restoration works in that scenario.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestIdxTogether{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20)),  # No db_index=True
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AlterIndexTogether(
                            name=model_name.lower(),
                            index_together={('a', 'b')},
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_idxtog',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Verify index_together index was restored
                self._assert_index_exists(
                    result.constraints,
                    expected_columns={'a', 'b'},
                    error_msg=(
                        f"index_together index on ('a', 'b') was not recreated after type change "
                        f"({self._get_context_description(use_single_migration)}). "
                        f"Expected index_together to be restored for field without db_index=True."
                    ),
                )

    @expectedFailure
    def test_unique_together_retained_when_field_also_has_unique_true(self):
        """
        Test that unique_together constraints are retained when a field with unique=True is altered.

        KNOWN BUG: When a field has BOTH unique=True AND participates in unique_together,
        only the single-field unique constraint is restored after field alteration.
        The unique_together constraint is NOT restored because the unique_together restoration is in an
        'else' block that only executes when the field does NOT have unique=True.

        Runs with both split and combined migrations
        """
        for use_single_migration in [False, True]:

            with self.subTest(single_migration=use_single_migration):
                suffix = '_combined' if use_single_migration else '_split'
                model_name = f'TestUniqueAndUniqTogether{suffix}'

                operations_a = [
                        migrations.CreateModel(
                            name=model_name,
                            fields=[
                                ('id', models.AutoField(primary_key=True)),
                                ('a', models.CharField(max_length=20, unique=True)),
                                ('b', models.CharField(max_length=20)),
                            ],
                        ),
                        migrations.AlterUniqueTogether(
                            name=model_name.lower(),
                            unique_together={('a', 'b')},
                        ),
                    ]

                operations_b = [
                        migrations.AlterField(
                            model_name=model_name.lower(),
                            name='a',
                            field=models.CharField(max_length=40, unique=True),
                        ),
                    ]

                result = self._run_migration_test(
                    operations_a=operations_a,
                    operations_b=operations_b,
                    migration_name_prefix='test_uniq_uniqtog',
                    model_name=model_name,
                    use_single_migration=use_single_migration,
                )

                # Check that single-field unique constraint on 'a' was restored
                single_unique_constraints = [
                    info for info in result.constraints.values()
                    if info.get('unique') and set(info['columns']) == {'a'}
                ]
                self.assertTrue(
                    len(single_unique_constraints) > 0,
                    f"Single-field unique constraint on 'a' was not restored "
                    f"({self._get_context_description(use_single_migration)})."
                )

                # Check that unique_together constraint on ('a', 'b') was restored
                # THIS ASSERTION WILL FAIL due to the bug in mssql/schema.py lines 838-871
                unique_together_constraints = [
                    info for info in result.constraints.values()
                    if info.get('unique') and set(info['columns']) == {'a', 'b'}
                ]
                self.assertTrue(
                    len(unique_together_constraints) > 0,
                    f"unique_together constraint on ('a', 'b') was not restored when field 'a' has unique=True "
                    f"({self._get_context_description(use_single_migration)}). "
                    f"This is a bug in mssql/schema.py: unique_together restoration is in an 'else' block "
                    f"that only executes when the field does NOT have unique=True."
                )





class TestAddAndAlterUniqueIndex(TestCase):

    def test_alter_unique_nullable_to_non_nullable(self):
        """
        Test a single migration that creates a field with unique=True and null=True and then alters
        the field to set null=False. See https://github.com/microsoft/mssql-django/issues/22
        """
        operations = [
            migrations.CreateModel(
                "TestAlterNullableInUniqueField",
                [
                    ("id", models.AutoField(primary_key=True)),
                    ("a", models.CharField(max_length=4, unique=True, null=True)),
                ]
            ),
            migrations.AlterField(
                "testalternullableinuniquefield",
                "a",
                models.CharField(max_length=4, unique=True)
            )
        ]

        project_state = ProjectState()
        new_state = project_state.clone()
        migration = Migration("name", "testapp")
        migration.operations = operations

        try:
            with connection.schema_editor(atomic=True) as editor:
                migration.apply(new_state, editor)
        except django.db.utils.ProgrammingError as e:
            self.fail('Check if can alter field from unique, nullable to unique non-nullable for issue #23, AlterField failed with exception: %s' % e)

class TestKeepIndexWithDbcomment(TestCase):
    def _find_key_with_type_idx(self, input_dict):
        for key, value in input_dict.items():
            if value.get("type") == "idx":
                return key
        return None

    @skipIf(VERSION < (4, 2), "db_comment not available before 4.2")
    def test_drop_foreignkey(self):
        app_label = "test_drop_foreignkey"
        operations = [
                migrations.CreateModel(
                    name="brand",
                    fields=[
                        ("id", models.AutoField(primary_key=True)),
                        ("name", models.CharField(max_length=100)),
                    ],
                ),
                migrations.CreateModel(
                    name="car1",
                    fields=[
                        ("id", models.AutoField(primary_key=True)),
                        (
                            "brand",
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                to="test_drop_foreignkey.brand",
                                related_name="car1",
                                db_constraint=True,
                            ),
                        ),
                    ],
                ),
                migrations.CreateModel(
                    name="car2",
                    fields=[
                        ("id", models.AutoField(primary_key=True)),
                        (
                            "brand",
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                to="test_drop_foreignkey.brand",
                                related_name="car2",
                                db_constraint=True,
                            ),
                        ),
                    ],
                ),
                migrations.CreateModel(
                    name="car3",
                    fields=[
                        ("id", models.AutoField(primary_key=True)),
                        (
                            "brand",
                            models.ForeignKey(
                                on_delete=django.db.models.deletion.CASCADE,
                                to="test_drop_foreignkey.brand",
                                related_name="car3",
                                db_constraint=True,
                            ),
                        ),
                    ],
                ),
            ]
        migration = Migration("name", app_label)
        migration.operations = operations
        with connection.schema_editor(atomic=True) as editor:
            project_state = migration.apply(ProjectState(), editor)

        alter_fk_car1 = migrations.AlterField(
            model_name="car1",
            name="brand",
            field=models.ForeignKey(
                to="test_drop_foreignkey.brand",
                on_delete=django.db.models.deletion.CASCADE,
                db_constraint=False,
                related_name="car1",
            ),
        )
        alter_fk_car2 = migrations.AlterField(
            model_name="car2",
            name="brand",
            field=models.ForeignKey(
                to="test_drop_foreignkey.brand",
                on_delete=django.db.models.deletion.CASCADE,
                db_constraint=False,
                related_name="car2",
                db_comment=""
            ),
        )
        alter_fk_car3 = migrations.AlterField(
            model_name="car3",
            name="brand",
            field=models.ForeignKey(
                to="test_drop_foreignkey.brand",
                on_delete=django.db.models.deletion.CASCADE,
                db_constraint=False,
                related_name="car3",
                db_comment="fk_on_delete_keep_index"
            ),
        )
        new_state = project_state.clone()
        with connection.schema_editor(atomic=True) as editor:
            alter_fk_car1.state_forwards("test_drop_foreignkey", new_state)
            alter_fk_car1.database_forwards(
                "test_drop_foreignkey", editor, project_state, new_state
            )
        car_index = self._find_key_with_type_idx(
            get_constraints(
                table_name=new_state.apps.get_model(
                    "test_drop_foreignkey", "car1"
                )._meta.db_table
            )
        )
        # Test alter foreignkey without db_comment field
        # The index should be dropped (keep the old behavior)
        self.assertIsNone(car_index)

        project_state = new_state
        new_state = new_state.clone()
        with connection.schema_editor(atomic=True) as editor:
            alter_fk_car2.state_forwards("test_drop_foreignkey", new_state)
            alter_fk_car2.database_forwards(
                "test_drop_foreignkey", editor, project_state, new_state
            )
        car_index = self._find_key_with_type_idx(
            get_constraints(
                table_name=new_state.apps.get_model(
                    "test_drop_foreignkey", "car2"
                )._meta.db_table
            )
        )
        # Test alter fk with empty db_comment
        self.assertIsNone(car_index)

        project_state = new_state
        new_state = new_state.clone()
        with connection.schema_editor(atomic=True) as editor:
            alter_fk_car3.state_forwards("test_drop_foreignkey", new_state)
            alter_fk_car3.database_forwards(
                "test_drop_foreignkey", editor, project_state, new_state
            )
        car_index = self._find_key_with_type_idx(
            get_constraints(
                table_name=new_state.apps.get_model(
                    "test_drop_foreignkey", "car3"
                )._meta.db_table
            )
        )
        # Test alter fk with fk_on_delete_keep_index in db_comment
        # Index should be preserved in this case
        self.assertIsNotNone(car_index)



# Copyright (c) Microsoft Corporation.
# Licensed under the BSD license.

import types
from functools import partial
from itertools import chain

import django
from django.db.models.aggregates import Avg, Count, StdDev, Variance
from django.db.models import AutoField
if django.VERSION >= (6, 0):
    from django.db.models.aggregates import StringAgg
from django.db.models.expressions import Col, OuterRef, Ref, Subquery, Value, Window
from django.db.models.expressions import ResolvedOuterRef
from django.db.models.functions import (
    Chr, ConcatPair, Greatest, Least, Length, LPad, Random, Repeat, RPad, StrIndex, Substr, Trim
)
from django.db.models.sql import compiler
from django.db.transaction import TransactionManagementError
from django.db.utils import NotSupportedError
if django.VERSION >= (3, 1):
    from django.db.models.fields.json import KeyTransform as json_KeyTransform
    # compile_json_path was moved to connection.ops in Django 6.0
    if django.VERSION < (6, 0):
        from django.db.models.fields.json import compile_json_path
    else:
        compile_json_path = None
if django.VERSION >= (4, 2):
    from django.core.exceptions import EmptyResultSet, FullResultSet
# ColPairs was introduced in Django 5.2 for composite primary key support.
# When an OrderBy wraps a ColPairs, Django's OrderBy.as_sql() joins all
# columns into a single comma-separated SQL string. We expand these at
# the expression level in get_order_by() so each ORDER BY item is always
# a single column expression.
if django.VERSION >= (5, 2):
    from django.db.models.expressions import ColPairs
else:
    ColPairs = None

def _as_sql_agv(self, compiler, connection):
    return self.as_sql(compiler, connection, template='%(function)s(CONVERT(float, %(field)s))')


def _as_sql_chr(self, compiler, connection):
    return self.as_sql(compiler, connection, function='NCHAR')


def _as_sql_concatpair(self, compiler, connection):
    if connection.sql_server_version < 2012:
        node = self.coalesce()
        return node.as_sql(compiler, connection, arg_joiner=' + ', template='%(expressions)s')
    else:
        return self.as_sql(compiler, connection)


def _as_sql_count(self, compiler, connection):
    return self.as_sql(compiler, connection, function='COUNT_BIG')


def _as_sql_greatest(self, compiler, connection):
    # SQL Server does not provide GREATEST function,
    # so we emulate it with a table value constructor
    # https://msdn.microsoft.com/en-us/library/dd776382.aspx
    template = '(SELECT MAX(value) FROM (VALUES (%(expressions)s)) AS _%(function)s(value))'
    return self.as_sql(compiler, connection, arg_joiner='), (', template=template)

def _as_sql_json_keytransform(self, compiler, connection):
    lhs, params, key_transforms = self.preprocess_lhs(compiler, connection)
    # Always prefer backend compilation when available so SQL Server-specific
    # escaping rules are applied consistently across Django versions.
    if hasattr(connection.ops, 'compile_json_path'):
        json_path = connection.ops.compile_json_path(key_transforms)
    else:
        json_path = compile_json_path(key_transforms)
    json_path = json_path.replace("'", "''")
    return (
        "COALESCE(JSON_QUERY(%s, '%s'), JSON_VALUE(%s, '%s'))" %
        ((lhs, json_path) * 2)
    ), tuple(params) * 2

def _as_sql_least(self, compiler, connection):
    # SQL Server does not provide LEAST function,
    # so we emulate it with a table value constructor
    # https://msdn.microsoft.com/en-us/library/dd776382.aspx
    template = '(SELECT MIN(value) FROM (VALUES (%(expressions)s)) AS _%(function)s(value))'
    return self.as_sql(compiler, connection, arg_joiner='), (', template=template)


def _as_sql_length(self, compiler, connection):
    return self.as_sql(compiler, connection, function='LEN')


def _as_sql_lpad(self, compiler, connection):
    i = iter(self.get_source_expressions())
    expression, expression_arg = compiler.compile(next(i))
    length, length_arg = compiler.compile(next(i))
    fill_text, fill_text_arg = compiler.compile(next(i))
    params = []
    params.extend(fill_text_arg)
    params.extend(length_arg)
    params.extend(length_arg)
    params.extend(expression_arg)
    params.extend(length_arg)
    params.extend(expression_arg)
    params.extend(expression_arg)
    params.extend(length_arg)
    template = ('LEFT(LEFT(REPLICATE(%(fill_text)s, %(length)s), CASE WHEN %(length)s > LEN(%(expression)s) '
                'THEN %(length)s - LEN(%(expression)s) ELSE 0 END) + %(expression)s, %(length)s)')
    return template % {'expression': expression, 'length': length, 'fill_text': fill_text}, params


def _as_sql_repeat(self, compiler, connection):
    return self.as_sql(compiler, connection, function='REPLICATE')


def _as_sql_rpad(self, compiler, connection):
    i = iter(self.get_source_expressions())
    expression, expression_arg = compiler.compile(next(i))
    length, length_arg = compiler.compile(next(i))
    fill_text, fill_text_arg = compiler.compile(next(i))
    params = []
    params.extend(expression_arg)
    params.extend(fill_text_arg)
    params.extend(length_arg)
    params.extend(length_arg)
    template = 'LEFT(%(expression)s + REPLICATE(%(fill_text)s, %(length)s), %(length)s)'
    return template % {'expression': expression, 'length': length, 'fill_text': fill_text}, params


def _as_sql_stddev(self, compiler, connection):
    function = 'STDEV'
    if self.function == 'STDDEV_POP':
        function = '%sP' % function
    return self.as_sql(compiler, connection, function=function)


def _as_sql_strindex(self, compiler, connection):
    self.source_expressions.reverse()
    sql = self.as_sql(compiler, connection, function='CHARINDEX')
    self.source_expressions.reverse()
    return sql


def _as_sql_substr(self, compiler, connection):
    if len(self.get_source_expressions()) < 3:
        self.get_source_expressions().append(Value(2**31 - 1))
    return self.as_sql(compiler, connection)


def _as_sql_trim(self, compiler, connection):
    return self.as_sql(compiler, connection, template='LTRIM(RTRIM(%(expressions)s))')


def _as_sql_variance(self, compiler, connection):
    function = 'VAR'
    if self.function == 'VAR_POP':
        function = '%sP' % function
    return self.as_sql(compiler, connection, function=function)


def _as_sql_stringagg(self, compiler, connection):
    if self.order_by and _contains_outerref(self.order_by, compiler.query):
        node = self.copy()
        node.order_by = None
        return node.as_sql(compiler, connection)

    template = None
    if self.order_by:
        template = '%(function)s(%(distinct)s%(expressions)s) WITHIN GROUP (%(order_by)s)%(filter)s'
    return self.as_sql(compiler, connection, template=template)


def _contains_outerref(expression, query=None):
    if expression is None:
        return False
    if isinstance(expression, (OuterRef, ResolvedOuterRef)):
        return True
    if isinstance(expression, Col) and query is not None:
        query_aliases = set(query.alias_map) if getattr(query, 'alias_map', None) else set()
        if expression.alias not in query_aliases:
            return True

    source_expressions = getattr(expression, 'get_source_expressions', None)
    if source_expressions is None:
        return False

    return any(_contains_outerref(expr, query) for expr in expression.get_source_expressions() if expr is not None)


def _as_sql_window(self, compiler, connection, template=None):
    # Get the expressions supported by the backend
    connection.ops.check_expression_support(self)
    # Raise an error if window expressions are not supported.
    if not connection.features.supports_over_clause:
        raise NotSupportedError("This backend does not support window expressions.")
    # Compile the source expression for the window function.
    expr_sql, params = compiler.compile(self.source_expression)
    # Initialize window SQL parts and parameters.
    window_sql, window_params = [], ()
    # Handle PARTITION BY clause if present.
    if self.partition_by is not None:
        # Compile the PARTITION BY clause.
        sql_expr, sql_params = self.partition_by.as_sql(
            compiler=compiler,
            connection=connection,
            template="PARTITION BY %(expressions)s",
        )
        window_sql.append(sql_expr)
        window_params += tuple(sql_params)

    # Handle ORDER BY clause if present.
    if self.order_by is not None:
        # Compile the ORDER BY clause.
        order_sql, order_params = compiler.compile(self.order_by)
        # Handles cases where order_by compiles to empty
        if not order_sql.strip():
            order_sql = "ORDER BY (SELECT NULL)"
            order_params = ()
        window_sql.append(order_sql)
        window_params += tuple(order_params)
    else:
        # Default to ORDER BY (SELECT NULL) if no order_by is specified.
        window_sql.append('ORDER BY (SELECT NULL)')

    # Handle frame specification if present.
    if self.frame:
        # Compile the frame clause.
        frame_sql, frame_params = compiler.compile(self.frame)
        window_sql.append(frame_sql)
        window_params += tuple(frame_params)

    # Use provided template or default to self.template.
    template = template or self.template
    # Return the formatted SQL and combined parameters.
    return (
        template % {"expression": expr_sql, "window": " ".join(window_sql).strip()},
        (*params, *window_params),
    )

def _cursor_iter(cursor, sentinel, col_count, itersize):
    """
    Yields blocks of rows from a cursor and ensures the cursor is closed when
    done.
    """
    if not hasattr(cursor.db, 'supports_mars') or cursor.db.supports_mars:
        # same as the original Django implementation
        try:
            for rows in iter((lambda: cursor.fetchmany(itersize)), sentinel):
                yield rows if col_count is None else [r[:col_count] for r in rows]
        finally:
            cursor.close()
    else:
        # retrieve all chunks from the cursor and close it before yielding
        # so that we can open an another cursor over an iteration
        # (for drivers such as FreeTDS)
        chunks = []
        try:
            for rows in iter((lambda: cursor.fetchmany(itersize)), sentinel):
                chunks.append(rows if col_count is None else [r[:col_count] for r in rows])
        finally:
            cursor.close()
        for rows in chunks:
            yield rows


compiler.cursor_iter = _cursor_iter


class SQLCompiler(compiler.SQLCompiler):

    def _resolve_order_by_source_expression(self, expression, dereference_ref=True):
        if expression is None:
            return None
        source_expressions = expression.get_source_expressions()
        if not source_expressions:
            return None
        source = source_expressions[0]
        if dereference_ref and isinstance(source, Ref):
            ref_source_expressions = source.get_source_expressions()
            if ref_source_expressions:
                source = ref_source_expressions[0]
        return source

    def _is_constant_order_by_expression(self, expression):
        if django.VERSION >= (4, 2):
            unresolved = self._resolve_order_by_source_expression(expression, dereference_ref=False)
            if isinstance(unresolved, Ref):
                return False
        source = self._resolve_order_by_source_expression(expression)
        return source is not None and self._is_constant_expression(source)

    def get_order_by(self):
        """
        Expand ColPairs-based OrderBy expressions into individual per-column
        OrderBy entries before SQL compilation.

        Django 5.2+ composite PKs produce OrderBy(ColPairs(...)) which
        compiles to a single comma-separated SQL string. SQL Server doesn't
        allow duplicate columns in ORDER BY (error 169), and expanding at
        the expression level lets us deduplicate individual columns cleanly
        without parsing SQL strings.
        """
        result = super().get_order_by()
        if ColPairs is None:
            return result
        expanded = []
        for resolved, (sql, params, is_ref) in result:
            if isinstance(getattr(resolved, 'expression', None), ColPairs):
                for col in resolved.expression.get_cols():
                    order = resolved.copy()
                    order.set_source_expressions([col])
                    col_sql, col_params = self.compile(order)
                    expanded.append((order, (col_sql, col_params, is_ref)))
            else:
                expanded.append((resolved, (sql, params, is_ref)))
        return expanded

    def as_sql(self, with_limits=True, with_col_aliases=False):
        """
        Create the SQL for this query. Return the SQL string and list of
        parameters.

        If 'with_limits' is False, any limit/offset information is not included
        in the query.
        """
        refcounts_before = self.query.alias_refcount.copy()
        try:
            extra_select, order_by, group_by = self.pre_sql_setup()
            for_update_part = None
            # Is a LIMIT/OFFSET clause needed?
            with_limit_offset = with_limits and (self.query.high_mark is not None or self.query.low_mark)
            combinator = self.query.combinator
            features = self.connection.features

            # The do_offset flag indicates whether we need to construct
            # the SQL needed to use limit/offset w/SQL Server.
            high_mark = self.query.high_mark
            low_mark = self.query.low_mark
            do_limit = with_limits and high_mark is not None
            do_offset = with_limits and low_mark != 0
            # SQL Server 2012 or newer supports OFFSET/FETCH clause
            supports_offset_clause = self.connection.sql_server_version >= 2012
            do_offset_emulation = do_offset and not supports_offset_clause

            if combinator:
                if not getattr(features, 'supports_select_{}'.format(combinator)):
                    raise NotSupportedError('{} is not supported on this database backend.'.format(combinator))
                result, params = self.get_combinator_sql(combinator, self.query.combinator_all)
            elif django.VERSION >= (4, 2) and self.qualify:
                result, params = self.get_qualify_sql()
                order_by = None
            else:
                distinct_fields, distinct_params = self.get_distinct()
                # This must come after 'select', 'ordering', and 'distinct' -- see
                # docstring of get_from_clause() for details.
                from_, f_params = self.get_from_clause()
                if django.VERSION >= (4, 2):
                    try:
                        where, w_params = self.compile(self.where) if self.where is not None else ("", [])
                    except EmptyResultSet:
                        if self.elide_empty:
                            raise
                        # Use a predicate that's always False.
                        where, w_params = "0 = 1", []
                    except FullResultSet:
                        where, w_params = "", []
                    try:
                        having, h_params = self.compile(self.having) if self.having is not None else ("", [])
                    except FullResultSet:
                        having, h_params = "", []
                else:
                    where, w_params = self.compile(self.where) if self.where is not None else ("", [])
                    having, h_params = self.compile(self.having) if self.having is not None else ("", [])
                params = []
                result = ['SELECT']

                if self.query.distinct:
                    distinct_result, distinct_params = self.connection.ops.distinct_sql(
                        distinct_fields,
                        distinct_params,
                    )
                    result += distinct_result
                    params += distinct_params

                # SQL Server requires the keword for limitting at the begenning
                if do_limit and not do_offset:
                    result.append('TOP %d' % high_mark)

                out_cols = []
                col_idx = 1
                for _, (s_sql, s_params), alias in self.select + extra_select:
                    if alias:
                        s_sql = '%s AS %s' % (s_sql, self.connection.ops.quote_name(alias))
                    elif with_col_aliases or do_offset_emulation:
                        s_sql = '%s AS %s' % (s_sql, 'Col%d' % col_idx)
                        col_idx += 1
                    params.extend(s_params)
                    out_cols.append(s_sql)

                # SQL Server requires an order-by clause for offsetting
                if do_offset:
                    meta = self.query.get_meta()
                    qn = self.quote_name_unless_alias
                    offsetting_order_by = '%s.%s' % (qn(meta.db_table), qn(meta.pk.db_column or meta.pk.column))
                    if do_offset_emulation:
                        if order_by:
                            ordering = []
                            seen_full = set()  # Full column refs (qualified or unqualified)
                            seen_unqualified = set()  # Just column names from unqualified refs
                            for expr, (o_sql, o_params, _) in order_by:
                                if self._is_constant_order_by_expression(expr):
                                    continue
                                # value_expression in OVER clause cannot refer to
                                # expressions or aliases in the select list. See:
                                # http://msdn.microsoft.com/en-us/library/ms189461.aspx
                                src = self._resolve_order_by_source_expression(expr, dereference_ref=False)
                                if isinstance(src, Ref):
                                    src = self._resolve_order_by_source_expression(expr)
                                    o_sql, _ = src.as_sql(self, self.connection)
                                    odir = 'DESC' if expr.descending else 'ASC'
                                    o_sql = '%s %s' % (o_sql, odir)
                                # SQL Server doesn't allow duplicate columns in ORDER BY.
                                # ColPairs are already expanded in get_order_by(), so each
                                # o_sql is a single column/expression.
                                # Handle the case where [col] and [table].[col] refer to
                                # the same column.
                                col_ref = o_sql.rsplit(' ', 1)[0] if o_sql.rstrip().endswith(('ASC', 'DESC')) else o_sql
                                col_ref_upper = col_ref.upper()
                                # Only treat as a qualified column ref if it looks like
                                # [table].[col], not a function call containing dots.
                                is_qualified = '.' in col_ref and '(' not in col_ref
                                col_name = col_ref.rsplit('.', 1)[-1].upper() if is_qualified else col_ref_upper
                                if col_ref_upper in seen_full:
                                    continue
                                if is_qualified and col_name in seen_unqualified:
                                    continue
                                seen_full.add(col_ref_upper)
                                if not is_qualified:
                                    seen_unqualified.add(col_name)
                                ordering.append(o_sql)
                                params.extend(o_params)
                            if ordering:
                                offsetting_order_by = ', '.join(ordering)
                            order_by = []
                        out_cols.append('ROW_NUMBER() OVER (ORDER BY %s) AS [rn]' % offsetting_order_by)
                    elif not order_by:
                        order_by.append(((None, ('%s ASC' % offsetting_order_by, [], None))))

                if self.query.select_for_update and self.connection.features.has_select_for_update:
                    if self.connection.get_autocommit():
                        raise TransactionManagementError('select_for_update cannot be used outside of a transaction.')

                    if with_limit_offset and not self.connection.features.supports_select_for_update_with_limit:
                        raise NotSupportedError(
                            'LIMIT/OFFSET is not supported with '
                            'select_for_update on this database backend.'
                        )
                    nowait = self.query.select_for_update_nowait
                    skip_locked = self.query.select_for_update_skip_locked
                    of = self.query.select_for_update_of
                    # If it's a NOWAIT/SKIP LOCKED/OF query but the backend
                    # doesn't support it, raise NotSupportedError to prevent a
                    # possible deadlock.
                    if nowait and not self.connection.features.has_select_for_update_nowait:
                        raise NotSupportedError('NOWAIT is not supported on this database backend.')
                    elif skip_locked and not self.connection.features.has_select_for_update_skip_locked:
                        raise NotSupportedError('SKIP LOCKED is not supported on this database backend.')
                    elif of and not self.connection.features.has_select_for_update_of:
                        raise NotSupportedError('FOR UPDATE OF is not supported on this database backend.')
                    for_update_part = self.connection.ops.for_update_sql(
                        nowait=nowait,
                        skip_locked=skip_locked,
                        of=self.get_select_for_update_of_arguments(),
                    )

                if for_update_part and self.connection.features.for_update_after_from:
                    from_.insert(1, for_update_part)

                result += [', '.join(out_cols)]
                if from_:
                    result += ['FROM', *from_]
                params.extend(f_params)

                if where:
                    result.append('WHERE %s' % where)
                    params.extend(w_params)

                grouping = []
                for g_sql, g_params in group_by:
                    grouping.append(g_sql)
                    params.extend(g_params)
                if grouping:
                    if distinct_fields:
                        raise NotImplementedError('annotate() + distinct(fields) is not implemented.')
                    if self.query.default_ordering and not self.query.order_by:
                        order_by = self.connection.ops.force_no_ordering()
                    else:
                        order_by = order_by or self.connection.ops.force_no_ordering()
                    result.append('GROUP BY %s' % ', '.join(grouping))

                if having:
                    result.append('HAVING %s' % having)
                    params.extend(h_params)

            explain = self.query.explain_info if django.VERSION >= (4, 0) else self.query.explain_query
            if explain:
                result.insert(0, self.connection.ops.explain_query_prefix(
                    self.query.explain_format,
                    **self.query.explain_options
                ))

            if order_by:
                ordering = []
                seen_full = set()  # Full column refs (qualified or unqualified)
                seen_unqualified = set()  # Just column names from unqualified refs
                for expr, (o_sql, o_params, _) in order_by:
                    json_key_transform_ordering = False
                    uses_ref_alias = False
                    # Build one or more ORDER BY items for this expression,
                    # then run all of them through the shared de-duplication
                    # logic below.
                    normalized_order_items = None
                    if self._is_constant_order_by_expression(expr):
                        continue
                    if expr:
                        unresolved_src = self._resolve_order_by_source_expression(
                            expr,
                            dereference_ref=False,
                        )
                        uses_ref_alias = isinstance(unresolved_src, Ref)
                        src = self._resolve_order_by_source_expression(expr)
                        if isinstance(src, Random):
                            # ORDER BY RAND() doesn't return rows in random order
                            # replace it with NEWID()
                            o_sql = o_sql.replace('RAND()', 'NEWID()')
                        elif isinstance(src, json_KeyTransform) and not uses_ref_alias:
                            json_key_transform_ordering = True
                    if json_key_transform_ordering:
                        direction = 'DESC' if getattr(expr, 'descending', False) else 'ASC'
                        stripped_o_sql = o_sql.strip()
                        if stripped_o_sql.upper().endswith(' DESC'):
                            base_o_sql = stripped_o_sql[:-5]
                        elif stripped_o_sql.upper().endswith(' ASC'):
                            base_o_sql = stripped_o_sql[:-4]
                        else:
                            base_o_sql = stripped_o_sql
                        if base_o_sql.isdigit():
                            json_key_transform_ordering = False
                        else:
                            # For JSON numeric ordering, use a numeric-first
                            # key and then a textual fallback key. Both keys
                            # must still go through the standard dedupe path.
                            normalized_order_items = [
                                ('TRY_CONVERT(float, %s) %s' % (base_o_sql, direction), o_params),
                                ('%s %s' % (base_o_sql, direction), o_params),
                            ]
                    if normalized_order_items is None:
                        # Default path: keep original ORDER BY SQL as-is.
                        normalized_order_items = [(o_sql, o_params)]
                    # SQL Server doesn't allow the same column to appear twice
                    # in ORDER BY. ColPairs are already expanded in
                    # get_order_by(), so each o_sql is a single expression.
                    # Handle the case where [col] and [table].[col] refer to
                    # the same column.
                    for normalized_sql, normalized_params in normalized_order_items:
                        # Normalize sort direction suffix so dedupe compares
                        # on the expression body, not ASC/DESC text noise.
                        col_ref = (
                            normalized_sql.rsplit(' ', 1)[0]
                            if normalized_sql.rstrip().endswith(('ASC', 'DESC'))
                            else normalized_sql
                        )
                        col_ref_upper = col_ref.upper()
                        is_qualified = '.' in col_ref and '(' not in col_ref
                        col_name = col_ref.rsplit('.', 1)[-1].upper() if is_qualified else col_ref_upper
                        if col_ref_upper in seen_full:
                            continue
                        if is_qualified and col_name in seen_unqualified:
                            continue
                        seen_full.add(col_ref_upper)
                        if not is_qualified:
                            seen_unqualified.add(col_name)
                        ordering.append(normalized_sql)
                        params.extend(normalized_params)
                if ordering:
                    result.append('ORDER BY %s' % ', '.join(ordering))
                else:
                    order_by = []
                    if do_offset and supports_offset_clause:
                        meta = self.query.get_meta()
                        qn = self.quote_name_unless_alias
                        result.append(
                            'ORDER BY %s.%s ASC' % (
                                qn(meta.db_table),
                                qn(meta.pk.db_column or meta.pk.column),
                            )
                        )

                # For subqueres with an ORDER BY clause, SQL Server also
                # requires a TOP or OFFSET clause which is not generated for
                # Django 2.x.  See https://github.com/microsoft/mssql-django/issues/12
                # Add OFFSET for all Django versions.
                # https://github.com/microsoft/mssql-django/issues/109
                if ordering and not (do_offset or do_limit) and supports_offset_clause:
                    result.append("OFFSET 0 ROWS")

            # SQL Server requires the backend-specific emulation (2008 or earlier)
            # or an offset clause (2012 or newer) for offsetting
            if do_offset:
                if do_offset_emulation:
                    # Construct the final SQL clause, using the initial select SQL
                    # obtained above.
                    result = ['SELECT * FROM (%s) AS X WHERE X.rn' % ' '.join(result)]
                    # Place WHERE condition on `rn` for the desired range.
                    if do_limit:
                        result.append('BETWEEN %d AND %d' % (low_mark + 1, high_mark))
                    else:
                        result.append('>= %d' % (low_mark + 1))
                    if not self.query.subquery:
                        result.append('ORDER BY X.rn')
                else:
                    result.append(self.connection.ops.limit_offset_sql(self.query.low_mark, self.query.high_mark))

            if self.query.subquery and extra_select:
                # If the query is used as a subquery, the extra selects would
                # result in more columns than the left-hand side expression is
                # expecting. This can happen when a subquery uses a combination
                # of order_by() and distinct(), forcing the ordering expressions
                # to be selected as well. Wrap the query in another subquery
                # to exclude extraneous selects.
                sub_selects = []
                sub_params = []
                for index, (select, _, alias) in enumerate(self.select, start=1):
                    if not alias and with_col_aliases:
                        alias = 'col%d' % index
                    if alias:
                        sub_selects.append("%s.%s" % (
                            self.connection.ops.quote_name('subquery'),
                            self.connection.ops.quote_name(alias),
                        ))
                    else:
                        select_clone = select.relabeled_clone({select.alias: 'subquery'})
                        subselect, subparams = select_clone.as_sql(self, self.connection)
                        sub_selects.append(subselect)
                        sub_params.extend(subparams)
                return 'SELECT %s FROM (%s) subquery' % (
                    ', '.join(sub_selects),
                    ' '.join(result),
                ), tuple(sub_params + params)

            return ' '.join(result), tuple(params)
        finally:
            # Finally do cleanup - get rid of the joins we created above.
            self.query.reset_refcounts(refcounts_before)

    def compile(self, node, *args, **kwargs):
        node = self._as_microsoft(node)
        return super().compile(node, *args, **kwargs)

    def collapse_group_by(self, expressions, having):
        expressions = super().collapse_group_by(expressions, having)
        # SQL server does not allow subqueries or constant expressions in the group by
        # For constants: Each GROUP BY expression must contain at least one column that is not an outer reference.
        # For subqueries: Cannot use an aggregate or a subquery in an expression used for the group by list of a GROUP BY clause.
        return self._filter_subquery_and_constant_expressions(expressions)

    def _is_constant_expression(self, expression):
        if expression is None:
            return False
        if isinstance(expression, Value):
            return True
        if not hasattr(expression, 'get_source_expressions'):
            return False
        sub_exprs = expression.get_source_expressions()
        if not sub_exprs:
            return False
        for each in sub_exprs:
            if each is None:
                return False
            if not self._is_constant_expression(each):
                return False
        return True



    def _filter_subquery_and_constant_expressions(self, expressions):
        ret = []
        for expression in expressions:
            if self._is_subquery(expression):
                continue
            if self._is_constant_expression(expression):
                continue
            if not self._has_nested_subquery(expression):
                ret.append(expression)
        return ret

    def _has_nested_subquery(self, expression):
        if self._is_subquery(expression):
            return True
        for sub_expr in expression.get_source_expressions():
            if self._has_nested_subquery(sub_expr):
                return True
        return False

    def _is_subquery(self, expression):
        return isinstance(expression, Subquery)

    def _as_microsoft(self, node):
        as_microsoft = None
        if isinstance(node, Avg):
            as_microsoft = _as_sql_agv
        elif isinstance(node, Chr):
            as_microsoft = _as_sql_chr
        elif isinstance(node, ConcatPair):
            as_microsoft = _as_sql_concatpair
        elif isinstance(node, Count):
            as_microsoft = _as_sql_count
        elif isinstance(node, Greatest):
            as_microsoft = _as_sql_greatest
        elif isinstance(node, Least):
            as_microsoft = _as_sql_least
        elif isinstance(node, Length):
            as_microsoft = _as_sql_length
        elif isinstance(node, RPad):
            as_microsoft = _as_sql_rpad
        elif isinstance(node, LPad):
            as_microsoft = _as_sql_lpad
        elif isinstance(node, Repeat):
            as_microsoft = _as_sql_repeat
        elif isinstance(node, StdDev):
            as_microsoft = _as_sql_stddev
        elif isinstance(node, StrIndex):
            as_microsoft = _as_sql_strindex
        elif isinstance(node, Substr):
            as_microsoft = _as_sql_substr
        elif isinstance(node, Trim):
            as_microsoft = _as_sql_trim
        elif isinstance(node, Variance):
            as_microsoft = _as_sql_variance
        elif django.VERSION >= (6, 0) and isinstance(node, StringAgg):
            as_microsoft = _as_sql_stringagg
        if django.VERSION >= (3, 1):
            if isinstance(node, json_KeyTransform):
                as_microsoft = _as_sql_json_keytransform
        if django.VERSION >= (4, 1):
            if isinstance(node, Window):
                as_microsoft = _as_sql_window
        if as_microsoft:
            node = node.copy()
            node.as_microsoft = types.MethodType(as_microsoft, node)
        return node


class SQLInsertCompiler(compiler.SQLInsertCompiler, SQLCompiler):
    def get_returned_fields(self):
        if django.VERSION >= (3, 0, 0):
            return self.returning_fields
        return self.return_id

    def can_return_columns_from_insert(self):
        if django.VERSION >= (3, 0, 0):
            return self.connection.features.can_return_columns_from_insert
        return self.connection.features.can_return_id_from_insert

    def can_return_rows_from_bulk_insert(self):
        if django.VERSION >= (3, 0, 0):
            return self.connection.features.can_return_rows_from_bulk_insert
        return self.connection.features.can_return_ids_from_bulk_insert

    def fix_auto(self, sql, opts, fields, qn):
        if opts.auto_field is not None:
            # db_column is None if not explicitly specified by model field
            auto_field_column = opts.auto_field.db_column or opts.auto_field.column
            columns = [f.column for f in fields]
            if auto_field_column in columns:
                id_insert_sql = []
                table = qn(opts.db_table)
                sql_format = 'SET IDENTITY_INSERT %s ON; %s; SET IDENTITY_INSERT %s OFF'
                for q, p in sql:
                    id_insert_sql.append((sql_format % (table, q, table), p))
                sql = id_insert_sql

        return sql

    def bulk_insert_default_values_sql(self, table):
        seed_rows_number = 8
        cross_join_power = 4  # 8^4 = 4096 > maximum allowed batch size for the backend = 1000

        def generate_seed_rows(n):
            return " UNION ALL ".join("SELECT 1 AS x" for _ in range(n))

        def cross_join(p):
            return ", ".join("SEED_ROWS AS _%s" % i for i in range(p))

        return """
        WITH SEED_ROWS AS (%s)
            MERGE INTO %s
            USING (
                SELECT TOP %s * FROM (SELECT 1 as x FROM %s) FAKE_ROWS
            ) FAKE_DATA
            ON 1 = 0
            WHEN NOT MATCHED THEN
            INSERT DEFAULT VALUES
        """ % (generate_seed_rows(seed_rows_number),
               table,
               len(self.query.objs),
               cross_join(cross_join_power))

    def as_sql(self):
        # We don't need quote_name_unless_alias() here, since these are all
        # going to be column names (so we can avoid the extra overhead).
        qn = self.connection.ops.quote_name
        opts = self.query.get_meta()
        result = ['INSERT INTO %s' % qn(opts.db_table)]

        if self.query.fields:
            fields = list(self.query.fields)
            supports_default_keyword_in_bulk_insert = (
                self.connection.features.supports_default_keyword_in_bulk_insert
            )
            result.append('(%s)' % ', '.join(qn(f.column) for f in fields))
            values_format = 'VALUES (%s)'

            if django.VERSION < (6, 0):
                value_rows = [
                    [self.prepare_value(field, self.pre_save_val(field, obj)) for field in fields]
                    for obj in self.query.objs
                ]
            else:
                from django.db.models.expressions import DatabaseDefault

                value_cols = []
                for field in list(fields):
                    field_prepare = partial(self.prepare_value, field)
                    field_pre_save = partial(self.pre_save_val, field)
                    field_values = [
                        field_prepare(field_pre_save(obj)) for obj in self.query.objs
                    ]

                    if not field.has_db_default():
                        value_cols.append(field_values)
                        continue

                    if len(fields) > 1 and all(
                        isinstance(value, DatabaseDefault) for value in field_values
                    ):
                        fields.remove(field)
                        continue

                    if supports_default_keyword_in_bulk_insert:
                        value_cols.append(field_values)
                        continue

                    prepared_db_default = field_prepare(field.db_default)
                    field_values = [
                        prepared_db_default
                        if isinstance(value, DatabaseDefault)
                        else value
                        for value in field_values
                    ]
                    value_cols.append(field_values)
                value_rows = list(zip(*value_cols))
                result[-1] = '(%s)' % ', '.join(qn(f.column) for f in fields)
        else:
            values_format = '%s VALUES'
            # An empty object.
            value_rows = [[self.connection.ops.pk_default_value()] for _ in self.query.objs]
            fields = [None]

        # Currently the backends just accept values when generating bulk
        # queries and generate their own placeholders. Doing that isn't
        # necessary and it should be possible to use placeholders and
        # expressions in bulk inserts too.
        can_bulk = (not self.get_returned_fields() and self.connection.features.has_bulk_insert) and self.query.fields

        placeholder_rows, param_rows = self.assemble_as_sql(fields, value_rows)

        if self.get_returned_fields() and self.can_return_columns_from_insert():
            if self.can_return_rows_from_bulk_insert():
                if not(self.query.fields):
                    # There isn't really a single statement to bulk multiple DEFAULT VALUES insertions,
                    # so we have to use a workaround:
                    # https://dba.stackexchange.com/questions/254771/insert-multiple-rows-into-a-table-with-only-an-identity-column
                    result = [self.bulk_insert_default_values_sql(qn(opts.db_table))]
                    r_sql, self.returning_params = self.connection.ops.return_insert_columns(self.get_returned_fields())
                    if r_sql:
                        result.append(r_sql)
                    sql = " ".join(result) + ";"
                    return [(sql, None)]
                # Regular bulk insert
                params = []
                r_sql, self.returning_params = self.connection.ops.return_insert_columns(self.get_returned_fields())
                if r_sql:
                    result.append(r_sql)
                    params += [self.returning_params]
                params += param_rows
                result.append(self.connection.ops.bulk_insert_sql(fields, placeholder_rows))
            else:
                returned_fields = self.get_returned_fields()
                use_scope_identity = (
                    len(returned_fields) == 1 and isinstance(returned_fields[0], AutoField)
                )

                if use_scope_identity:
                    result.insert(0, 'SET NOCOUNT ON')
                    if not self.query.fields:
                        result.append('DEFAULT VALUES;')
                        params = []
                    else:
                        result.append((values_format + ';') % ', '.join(placeholder_rows[0]))
                        params = [param_rows[0]]
                    result.append('SELECT CAST(SCOPE_IDENTITY() AS bigint)')
                else:
                    params = []
                    table_name = qn(opts.db_table)
                    tmp_table_name = '#django_returning_insert'
                    returned_columns = ', '.join(qn(field.column) for field in returned_fields)
                    select_into_columns = []
                    for field in returned_fields:
                        column_sql = qn(field.column)
                        if isinstance(field, AutoField):
                            select_into_columns.append(f'CAST({column_sql} AS bigint) AS {column_sql}')
                        else:
                            select_into_columns.append(column_sql)

                    r_sql, self.returning_params = self.connection.ops.return_insert_columns(returned_fields)
                    if r_sql and self.returning_params:
                        params.append(self.returning_params)

                    insert_sql = result[:]
                    if r_sql:
                        insert_sql.append(f'{r_sql} INTO {tmp_table_name}')
                    if not self.query.fields:
                        insert_sql.append('DEFAULT VALUES')
                    else:
                        insert_sql.append(values_format % ', '.join(placeholder_rows[0]))
                        params.append(param_rows[0])

                    sql_batch = '; '.join([
                        'SET NOCOUNT ON',
                        f"IF OBJECT_ID('tempdb..{tmp_table_name}') IS NOT NULL DROP TABLE {tmp_table_name}",
                        f"SELECT TOP 0 {', '.join(select_into_columns)} INTO {tmp_table_name} FROM {table_name}",
                        ' '.join(insert_sql),
                        f'SELECT {returned_columns} FROM {tmp_table_name}',
                        f'DROP TABLE {tmp_table_name}',
                    ])
                    sql = [(sql_batch, tuple(chain.from_iterable(params)))]
                    if self.query.fields:
                        sql = self.fix_auto(sql, opts, fields, qn)
                    return sql
            sql = [(" ".join(result), tuple(chain.from_iterable(params)))]
        else:
            if can_bulk:
                result.append(self.connection.ops.bulk_insert_sql(fields, placeholder_rows))
                sql = [(" ".join(result), tuple(p for ps in param_rows for p in ps))]
            else:
                sql = [
                    (" ".join(result + [values_format % ", ".join(p)]), vals)
                    for p, vals in zip(placeholder_rows, param_rows)
                ]

        if self.query.fields:
            sql = self.fix_auto(sql, opts, fields, qn)

        return sql


class SQLDeleteCompiler(compiler.SQLDeleteCompiler, SQLCompiler):
    def as_sql(self):
        sql, params = super().as_sql()
        if sql:
            sql = '; '.join(['SET NOCOUNT OFF', sql])
        return sql, params


class SQLUpdateCompiler(compiler.SQLUpdateCompiler, SQLCompiler):
    def as_sql(self):
        sql, params = super().as_sql()
        if sql:
            sql = '; '.join(['SET NOCOUNT OFF', sql])
        return sql, params


class SQLAggregateCompiler(compiler.SQLAggregateCompiler, SQLCompiler):
    pass



# Copyright (c) Microsoft Corporation.
# Licensed under the BSD license.

"""
Tests for mssql/base.py utility functions and classes.
"""
import struct
import datetime
from decimal import Decimal
from uuid import UUID
from unittest import mock

from django.test import TestCase, SimpleTestCase

from mssql.base import (
    encode_connection_string,
    encode_value,
    prepare_token_for_odbc,
    handle_datetimeoffset,
    DatabaseWrapper,
)


class TestEncodeValue(SimpleTestCase):
    """Tests for the encode_value function."""

    def test_simple_value(self):
        """Simple values without special characters should pass through unchanged."""
        self.assertEqual(encode_value('simple'), 'simple')
        self.assertEqual(encode_value('MyPassword123'), 'MyPassword123')

    def test_value_with_semicolon(self):
        """Values containing semicolons should be wrapped in curly braces."""
        self.assertEqual(encode_value('pass;word'), '{pass;word}')
        self.assertEqual(encode_value('a;b;c'), '{a;b;c}')

    def test_value_starting_with_curly_brace(self):
        """Values starting with { should be wrapped and escaped."""
        # '{value}' -> the } is escaped to }}, then wrapped: {{value}}}
        self.assertEqual(encode_value('{value}'), '{{value}}}')
        self.assertEqual(encode_value('  {spaced}'), '{  {spaced}}}')

    def test_value_with_right_curly_brace(self):
        """Right curly braces should be escaped when wrapping is needed."""
        self.assertEqual(encode_value('pass}word;'), '{pass}}word;}')

    def test_empty_string(self):
        """Empty string should pass through unchanged."""
        self.assertEqual(encode_value(''), '')


class TestEncodeConnectionString(SimpleTestCase):
    """Tests for the encode_connection_string function."""

    def test_simple_connection_string(self):
        """Test basic connection string encoding."""
        fields = {'DRIVER': 'ODBC Driver 18 for SQL Server', 'SERVER': 'localhost'}
        result = encode_connection_string(fields)
        self.assertIn('DRIVER=ODBC Driver 18 for SQL Server', result)
        self.assertIn('SERVER=localhost', result)

    def test_connection_string_with_special_chars(self):
        """Test connection string with values containing special characters."""
        fields = {'PASSWORD': 'pass;word'}
        result = encode_connection_string(fields)
        self.assertEqual(result, 'PASSWORD={pass;word}')

    def test_empty_fields(self):
        """Test with empty fields dictionary."""
        result = encode_connection_string({})
        self.assertEqual(result, '')

    def test_multiple_fields(self):
        """Test with multiple fields."""
        fields = {
            'DRIVER': 'ODBC Driver 18 for SQL Server',
            'SERVER': 'localhost',
            'DATABASE': 'testdb',
            'UID': 'testuser',
        }
        result = encode_connection_string(fields)
        # All keys should be present
        for key in fields:
            self.assertIn(f'{key}=', result)


class TestPrepareTokenForOdbc(SimpleTestCase):
    """Tests for the prepare_token_for_odbc function."""

    def test_simple_token(self):
        """Test token preparation for a simple string."""
        token = "ABC"
        result = prepare_token_for_odbc(token)
        # Result should be bytes
        self.assertIsInstance(result, bytes)
        # Should start with 4-byte length header
        length = struct.unpack("=i", result[:4])[0]
        self.assertEqual(length, len(token) * 2)  # Each char becomes 2 bytes

    def test_token_content(self):
        """Test that token is properly expanded with null bytes."""
        token = "AB"
        result = prepare_token_for_odbc(token)
        # Skip the 4-byte length header
        payload = result[4:]
        # 'A' should be followed by null byte
        self.assertEqual(payload[0], ord('A'))
        self.assertEqual(payload[1], 0)
        # 'B' should be followed by null byte
        self.assertEqual(payload[2], ord('B'))
        self.assertEqual(payload[3], 0)

    def test_invalid_token_type(self):
        """Test that non-string tokens raise TypeError."""
        with self.assertRaises(TypeError) as cm:
            prepare_token_for_odbc(12345)
        self.assertIn("Invalid token format", str(cm.exception))

        with self.assertRaises(TypeError):
            prepare_token_for_odbc(b"bytes_token")

        with self.assertRaises(TypeError):
            prepare_token_for_odbc(None)

    def test_empty_token(self):
        """Test with empty string token."""
        token = ""
        result = prepare_token_for_odbc(token)
        # Should have 4-byte header with length 0
        length = struct.unpack("=i", result[:4])[0]
        self.assertEqual(length, 0)


class TestHandleDatetimeoffset(SimpleTestCase):
    """Tests for the handle_datetimeoffset function."""

    def test_datetime_conversion(self):
        """Test conversion of binary datetime offset to Python datetime."""
        # Pack a known datetime: 2023-06-15 14:30:45.123456
        # Format: year, month, day, hour, minute, second, nanoseconds (as microseconds * 1000), tz_hour, tz_min
        dto_bytes = struct.pack("<6hI2h", 2023, 6, 15, 14, 30, 45, 123456000, 0, 0)
        result = handle_datetimeoffset(dto_bytes)

        self.assertIsInstance(result, datetime.datetime)
        self.assertEqual(result.year, 2023)
        self.assertEqual(result.month, 6)
        self.assertEqual(result.day, 15)
        self.assertEqual(result.hour, 14)
        self.assertEqual(result.minute, 30)
        self.assertEqual(result.second, 45)
        self.assertEqual(result.microsecond, 123456)

    def test_datetime_edge_case(self):
        """Test with edge case values."""
        # Midnight on Jan 1, 2000
        dto_bytes = struct.pack("<6hI2h", 2000, 1, 1, 0, 0, 0, 0, 0, 0)
        result = handle_datetimeoffset(dto_bytes)

        self.assertEqual(result.year, 2000)
        self.assertEqual(result.month, 1)
        self.assertEqual(result.day, 1)
        self.assertEqual(result.hour, 0)
        self.assertEqual(result.minute, 0)
        self.assertEqual(result.second, 0)
        self.assertEqual(result.microsecond, 0)


class TestDatabaseWrapperIsDriverNotFoundError(SimpleTestCase):
    """Tests for the _is_driver_not_found_error method."""

    def setUp(self):
        # Create a mock DatabaseWrapper to test the method
        self.wrapper = object.__new__(DatabaseWrapper)

    def test_driver_not_found_libsodbc(self):
        """Test detection of 'can't open lib' error."""
        exception = Exception("Can't open lib 'ODBC Driver 18 for SQL Server'")
        self.assertTrue(self.wrapper._is_driver_not_found_error(exception))

    def test_driver_not_found_dsn(self):
        """Test detection of 'data source name not found' error."""
        exception = Exception("[IM002] Data source name not found")
        self.assertTrue(self.wrapper._is_driver_not_found_error(exception))

    def test_driver_not_found_explicit(self):
        """Test detection of 'driver not found' error."""
        exception = Exception("Driver not found: ODBC Driver 18")
        self.assertTrue(self.wrapper._is_driver_not_found_error(exception))

    def test_driver_could_not_be_loaded(self):
        """Test detection of 'specified driver could not be loaded' error."""
        exception = Exception("The specified driver could not be loaded")
        self.assertTrue(self.wrapper._is_driver_not_found_error(exception))

    def test_other_error(self):
        """Test that other errors are not detected as driver not found."""
        exception = Exception("Connection timeout expired")
        self.assertFalse(self.wrapper._is_driver_not_found_error(exception))

        exception = Exception("Login failed for user")
        self.assertFalse(self.wrapper._is_driver_not_found_error(exception))


class TestDatabaseWrapperBuildConnectionString(SimpleTestCase):
    """Tests for the _build_connection_string method."""

    def setUp(self):
        # Create a mock DatabaseWrapper
        self.wrapper = object.__new__(DatabaseWrapper)

    def test_basic_connection_string(self):
        """Test basic connection string building."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'localhost',
            'USER': 'testuser',
            'PASSWORD': 'testpass',
            'OPTIONS': {},
        }
        driver = 'ODBC Driver 18 for SQL Server'
        result = self.wrapper._build_connection_string(conn_params, driver)

        self.assertIn('DRIVER=ODBC Driver 18 for SQL Server', result)
        self.assertIn('SERVER=localhost', result)
        self.assertIn('DATABASE=testdb', result)
        self.assertIn('UID=testuser', result)
        self.assertIn('PWD=testpass', result)

    def test_connection_string_with_port(self):
        """Test connection string with port number."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'localhost',
            'PORT': 1433,
            'USER': 'testuser',
            'PASSWORD': 'testpass',
            'OPTIONS': {},
        }
        driver = 'ODBC Driver 18 for SQL Server'
        result = self.wrapper._build_connection_string(conn_params, driver)

        # Microsoft drivers use comma for port
        self.assertIn('SERVER=localhost,1433', result)

    def test_connection_string_with_dsn(self):
        """Test connection string with DSN."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'localhost',
            'OPTIONS': {'dsn': 'MyDSN'},
        }
        driver = 'ODBC Driver 18 for SQL Server'
        result = self.wrapper._build_connection_string(conn_params, driver)

        self.assertIn('DSN=MyDSN', result)
        # DRIVER should not be present when using DSN
        self.assertNotIn('DRIVER=', result)

    def test_connection_string_trusted_connection(self):
        """Test connection string with trusted connection (no user/password)."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'localhost',
            'OPTIONS': {},
        }
        driver = 'ODBC Driver 18 for SQL Server'
        result = self.wrapper._build_connection_string(conn_params, driver)

        self.assertIn('Trusted_Connection=yes', result)
        self.assertNotIn('UID=', result)
        self.assertNotIn('PWD=', result)

    def test_connection_string_with_extra_params(self):
        """Test connection string with extra parameters."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'localhost',
            'USER': 'testuser',
            'PASSWORD': 'testpass',
            'OPTIONS': {
                'extra_params': 'Encrypt=yes;TrustServerCertificate=yes'
            },
        }
        driver = 'ODBC Driver 18 for SQL Server'
        result = self.wrapper._build_connection_string(conn_params, driver)

        self.assertIn('Encrypt=yes', result)
        self.assertIn('TrustServerCertificate=yes', result)

    def test_connection_string_freetds(self):
        """Test connection string building for FreeTDS driver."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'myserver',
            'PORT': 1433,
            'OPTIONS': {
                'host_is_server': True,
            },
        }
        driver = 'FreeTDS'
        result = self.wrapper._build_connection_string(conn_params, driver)

        # FreeTDS uses PORT separately when host_is_server is True
        self.assertIn('SERVER=myserver', result)
        self.assertIn('PORT=1433', result)
        self.assertIn('Integrated Security=SSPI', result)

    def test_connection_string_active_directory_interactive(self):
        """Test that PASSWORD is not included with ActiveDirectoryInteractive auth."""
        conn_params = {
            'NAME': 'testdb',
            'HOST': 'localhost',
            'USER': 'user@domain.com',
            'PASSWORD': 'ignored',
            'OPTIONS': {
                'extra_params': 'Authentication=ActiveDirectoryInteractive'
            },
        }
        driver = 'ODBC Driver 18 for SQL Server'
        result = self.wrapper._build_connection_string(conn_params, driver)

        self.assertIn('UID=user@domain.com', result)
        self.assertNotIn('PWD=', result)


class TestCursorWrapperAsSqlType(SimpleTestCase):
    """Tests for CursorWrapper._as_sql_type method."""

    def setUp(self):
        from mssql.base import CursorWrapper
        # Create a mock CursorWrapper
        mock_cursor = mock.MagicMock()
        mock_connection = mock.MagicMock()
        mock_connection.driver_charset = None
        self.wrapper = CursorWrapper(mock_cursor, mock_connection)

    def test_string_types(self):
        """Test SQL type detection for strings."""
        self.assertEqual(self.wrapper._as_sql_type(str, ''), 'NVARCHAR')
        self.assertEqual(self.wrapper._as_sql_type(str, 'short'), 'NVARCHAR(5)')
        self.assertEqual(self.wrapper._as_sql_type(str, 'x' * 5000), 'NVARCHAR(max)')

    def test_integer_types(self):
        """Test SQL type detection for integers."""
        self.assertEqual(self.wrapper._as_sql_type(int, 100), 'INT')
        self.assertEqual(self.wrapper._as_sql_type(int, -100), 'INT')
        # Values exceeding INT range should be BIGINT
        self.assertEqual(self.wrapper._as_sql_type(int, 0x7FFFFFFF + 1), 'BIGINT')
        self.assertEqual(self.wrapper._as_sql_type(int, -0x7FFFFFFF - 1), 'BIGINT')

    def test_float_type(self):
        """Test SQL type detection for floats."""
        self.assertEqual(self.wrapper._as_sql_type(float, 3.14), 'DOUBLE PRECISION')

    def test_bool_type(self):
        """Test SQL type detection for booleans."""
        self.assertEqual(self.wrapper._as_sql_type(bool, True), 'BIT')
        self.assertEqual(self.wrapper._as_sql_type(bool, False), 'BIT')

    def test_decimal_type(self):
        """Test SQL type detection for Decimal."""
        self.assertEqual(self.wrapper._as_sql_type(Decimal, Decimal('123.45')), 'NUMERIC')

    def test_datetime_types(self):
        """Test SQL type detection for datetime types."""
        self.assertEqual(
            self.wrapper._as_sql_type(datetime.datetime, datetime.datetime.now()),
            'DATETIME2'
        )
        self.assertEqual(
            self.wrapper._as_sql_type(datetime.date, datetime.date.today()),
            'DATE'
        )
        self.assertEqual(
            self.wrapper._as_sql_type(datetime.time, datetime.time(12, 30)),
            'TIME'
        )

    def test_uuid_type(self):
        """Test SQL type detection for UUID."""
        self.assertEqual(
            self.wrapper._as_sql_type(UUID, UUID('12345678-1234-5678-1234-567812345678')),
            'uniqueidentifier'
        )

    def test_bytes_type(self):
        """Test SQL type detection for bytes."""
        self.assertEqual(self.wrapper._as_sql_type(bytes, b'binary_data'), 'VARBINARY')

    def test_unsupported_type(self):
        """Test that unsupported types raise NotImplementedError."""
        with self.assertRaises(NotImplementedError):
            self.wrapper._as_sql_type(list, [1, 2, 3])


class TestCursorWrapperFormatSql(SimpleTestCase):
    """Tests for CursorWrapper.format_sql method."""

    def setUp(self):
        from mssql.base import CursorWrapper
        mock_cursor = mock.MagicMock()
        mock_connection = mock.MagicMock()
        mock_connection.driver_charset = None
        self.wrapper = CursorWrapper(mock_cursor, mock_connection)

    def test_format_sql_no_params(self):
        """Test SQL formatting with no parameters."""
        sql = "SELECT * FROM users"
        result = self.wrapper.format_sql(sql, None)
        self.assertEqual(result, "SELECT * FROM users")

    def test_format_sql_empty_params(self):
        """Test SQL formatting with empty params list."""
        sql = "SELECT * FROM users"
        result = self.wrapper.format_sql(sql, [])
        self.assertEqual(result, "SELECT * FROM users")

    def test_format_sql_with_params(self):
        """Test SQL formatting replaces %s with ?."""
        sql = "SELECT * FROM users WHERE id = %s AND name = %s"
        result = self.wrapper.format_sql(sql, ['param1', 'param2'])
        self.assertEqual(result, "SELECT * FROM users WHERE id = ? AND name = ?")


class TestCursorWrapperFormatParams(SimpleTestCase):
    """Tests for CursorWrapper.format_params method."""

    def setUp(self):
        from mssql.base import CursorWrapper
        mock_cursor = mock.MagicMock()
        mock_connection = mock.MagicMock()
        mock_connection.driver_charset = None
        self.wrapper = CursorWrapper(mock_cursor, mock_connection)

    def test_format_params_none(self):
        """Test formatting with None params."""
        result = self.wrapper.format_params(None)
        self.assertEqual(result, ())

    def test_format_params_string(self):
        """Test formatting string parameters."""
        result = self.wrapper.format_params(['hello', 'world'])
        self.assertEqual(result, ('hello', 'world'))

    def test_format_params_bytes(self):
        """Test formatting bytes parameters."""
        result = self.wrapper.format_params([b'binary'])
        self.assertEqual(result, (b'binary',))

    def test_format_params_bool(self):
        """Test formatting boolean parameters (converted to 1/0)."""
        result = self.wrapper.format_params([True, False])
        self.assertEqual(result, (1, 0))

    def test_format_params_mixed(self):
        """Test formatting mixed parameter types."""
        result = self.wrapper.format_params([True, 'text', 123, None])
        self.assertEqual(result, (1, 'text', 123, None))

    def test_format_params_with_driver_charset(self):
        """Test formatting with driver charset encoding."""
        from mssql.base import CursorWrapper
        mock_cursor = mock.MagicMock()
        mock_connection = mock.MagicMock()
        mock_connection.driver_charset = 'utf-8'
        wrapper = CursorWrapper(mock_cursor, mock_connection)

        result = wrapper.format_params(['unicode: \u00e9'])
        # String should be encoded
        self.assertEqual(result[0], 'unicode: é')



# Copyright (c) Microsoft Corporation.
# Licensed under the BSD license.

import datetime
from unittest import skipUnless

from django import VERSION
from django.db.models import CharField, IntegerField, F
from django.db.models.expressions import Case, Exists, OuterRef, Subquery, Value, When, ExpressionWrapper
from django.test import TestCase, skipUnlessDBFeature

from django.db.models.aggregates import Count, Sum

if VERSION >= (6, 0):
    from django.db.models import StringAgg

from ..models import Author, Book, Comment, Post, Editor, ModelWithNullableFieldsOfDifferentTypes, Publisher


DJANGO3 = VERSION[0] >= 3


class TestSubquery(TestCase):
    def setUp(self):
        self.author = Author.objects.create(name="author")
        self.post = Post.objects.create(title="foo", author=self.author)

    def test_with_count(self):
        newest = Comment.objects.filter(post=OuterRef('pk')).order_by('-created_at')
        Post.objects.annotate(
            post_exists=Subquery(newest.values('text')[:1])
        ).filter(post_exists=True).count()


class TestExists(TestCase):
    def setUp(self):
        self.author = Author.objects.create(name="author")
        self.post = Post.objects.create(title="foo", author=self.author)

    def test_with_count(self):
        Post.objects.annotate(
            post_exists=Exists(Post.objects.all())
        ).filter(post_exists=True).count()

    @skipUnless(DJANGO3, "Django 3 specific tests")
    def test_with_case_when(self):
        author = Author.objects.annotate(
            has_post=Case(
                When(Exists(Post.objects.filter(author=OuterRef('pk')).values('pk')), then=Value(1)),
                default=Value(0),
                output_field=IntegerField(),
            )
        ).get()
        self.assertEqual(author.has_post, 1)

    def test_unnecessary_exists_group_by(self):
        author = Author.objects.annotate(
            has_post=Case(
                When(Exists(Post.objects.filter(author=OuterRef('pk')).values('pk')), then=Value(1)),
                default=Value(0),
                output_field=IntegerField(),
            )).annotate(
            amount=Count("post")
        ).get()
        self.assertEqual(author.amount, 1)
        self.assertEqual(author.has_post, 1)

    def test_combined_expression_annotation_with_aggregation(self):
        book = Author.objects.annotate(
            combined=ExpressionWrapper(
                Value(2) * Value(5), output_field=IntegerField()
            ),
            null_value=ExpressionWrapper(
                Value(None), output_field=IntegerField()
            ),
            rating_count=Count("post"),
        ).first()
        self.assertEqual(book.combined, 10)
        self.assertEqual(book.null_value, None)


    @skipUnless(DJANGO3, "Django 3 specific tests")
    def test_order_by_exists(self):
        author_without_posts = Author.objects.create(name="other author")
        authors_by_posts = Author.objects.order_by(Exists(Post.objects.filter(author=OuterRef('pk'))).desc())
        self.assertSequenceEqual(authors_by_posts, [self.author, author_without_posts])

        authors_by_posts = Author.objects.order_by(Exists(Post.objects.filter(author=OuterRef('pk'))).asc())
        self.assertSequenceEqual(authors_by_posts, [author_without_posts, self.author])


class TestGroupBy(TestCase):
    def test_group_by_case(self):
        annotated_queryset = Book.objects.annotate(age=Case(
            When(id__gt=1000, then=Value("new")),
            default=Value("old"),
            output_field=CharField())).values('age').annotate(sum=Sum('id'))
        self.assertEqual(list(annotated_queryset.all()), [])


class TestOrderingRegressions(TestCase):
    def setUp(self):
        Author.objects.bulk_create([
            Author(name='alice'),
            Author(name='bob'),
            Author(name='charlie'),
        ])

    def test_order_by_case_when_constant_value_executes(self):
        queryset = Author.objects.order_by(
            Case(
                When(name__isnull=False, then=Value(1)),
                default=Value(1),
                output_field=IntegerField(),
            )
        )
        self.assertCountEqual(
            list(queryset.values_list('name', flat=True)),
            ['alice', 'bob', 'charlie'],
        )

    def test_order_by_case_when_constant_value_with_offset_executes(self):
        queryset = Author.objects.order_by(Value(1))[1:3]
        expected = list(Author.objects.order_by('pk').values_list('name', flat=True))[1:3]
        self.assertEqual(
            list(queryset.values_list('name', flat=True)),
            expected,
        )


class TestModuloExpressionRegressions(TestCase):
    def test_modulo_expression_with_value_parameter_executes(self):
        author = Author.objects.create(name='mod-author')
        annotated = Author.objects.filter(pk=author.pk).annotate(
            mod_value=F('pk') % Value(2)
        ).values_list('mod_value', flat=True)
        self.assertEqual(list(annotated), [author.pk % 2])

@skipUnless(DJANGO3, "Django 3 specific tests")
@skipUnlessDBFeature("order_by_nulls_first")
class TestOrderBy(TestCase):
    def setUp(self):
        self.author = Author.objects.create(name="author")
        self.post = Post.objects.create(title="foo", author=self.author)
        self.editor = Editor.objects.create(name="editor")
        self.post_alt = Post.objects.create(title="Post with editor", author=self.author, alt_editor=self.editor)

    def test_order_by_nulls_last(self):
        results = Post.objects.order_by(F("alt_editor").asc(nulls_last=True)).all()
        self.assertEqual(len(results), 2)
        self.assertIsNotNone(results[0].alt_editor)
        self.assertIsNone(results[1].alt_editor)

    def test_order_by_nulls_first(self):
        results = Post.objects.order_by(F("alt_editor").desc(nulls_first=True)).all()
        self.assertEqual(len(results), 2)
        self.assertIsNone(results[0].alt_editor)
        self.assertIsNotNone(results[1].alt_editor)

class TestBulkUpdate(TestCase):
     def test_bulk_update_different_column_types(self):
        data = (
            (1, 'a', datetime.datetime(year=2024, month=1, day=1)),
            (2, 'b', datetime.datetime(year=2023, month=12, day=31))
        )
        objs = ModelWithNullableFieldsOfDifferentTypes.objects.bulk_create(ModelWithNullableFieldsOfDifferentTypes(int_value=row_data[0],
                                                                                                                   name=row_data[1],
                                                                                                                   date=row_data[2]) for row_data in data)
        for obj in objs:
            obj.int_value = None
            obj.name = None
            obj.date = None
        ModelWithNullableFieldsOfDifferentTypes.objects.bulk_update(objs, ["int_value", "name", "date"])
        self.assertCountEqual(ModelWithNullableFieldsOfDifferentTypes.objects.filter(int_value__isnull=True), objs)
        self.assertCountEqual(ModelWithNullableFieldsOfDifferentTypes.objects.filter(name__isnull=True), objs)
        self.assertCountEqual(ModelWithNullableFieldsOfDifferentTypes.objects.filter(date__isnull=True), objs)


class TestStringAggOrderingRegression(TestCase):
    @skipUnless(VERSION >= (6, 0), "StringAgg ordering is Django 6.0+")
    def test_stringagg_honors_ordering(self):
        Author.objects.bulk_create([
            Author(name='Charlie'),
            Author(name='Alice'),
            Author(name='Bob'),
        ])
        with self.assertNumQueries(1) as ctx:
            result = Author.objects.aggregate(
                names=StringAgg('name', delimiter=Value(', '), order_by=F('name'))
            )
        self.assertEqual(result['names'], 'Alice, Bob, Charlie')
        self.assertIn('WITHIN GROUP (', ctx[0]['sql'])
        self.assertIn('ORDER BY [testapp_author].[name]', ctx[0]['sql'])

    @skipUnless(VERSION >= (6, 0), "StringAgg ordering is Django 6.0+")
    def test_stringagg_order_by_outerref_does_not_use_within_group(self):
        publisher_1 = Publisher.objects.create(name='p1')
        Book.objects.create(name='Alpha', publisher=publisher_1)

        with self.assertNumQueries(1) as ctx:
            values = list(
                Publisher.objects.filter(pk=publisher_1.pk).annotate(
                    names=Subquery(
                        Book.objects.annotate(
                            names=StringAgg(
                                'name',
                                delimiter=Value(';'),
                                order_by=OuterRef('pk'),
                            )
                        ).values('names')[:1]
                    )
                ).values_list('names', flat=True)
            )

        self.assertEqual(values, ['Alpha'])
        self.assertNotIn('WITHIN GROUP', ctx[0]['sql'])



import unittest

import django.db.utils
from django import VERSION
from django.db import connections, connection, models
from django.db.models.functions import Now
from django.test import TransactionTestCase, TestCase, skipUnlessDBFeature
from django.utils import timezone

from ..models import Author, BinaryData


class TestTableWithTrigger(TransactionTestCase):
    def test_insert_into_table_with_trigger(self):
        connection = connections['default']
        with connection.schema_editor() as cursor:
            cursor.execute("""
                CREATE TRIGGER TestTrigger
                ON [testapp_author]
                FOR INSERT
                AS
                INSERT INTO [testapp_editor]([name]) VALUES ('Bar')
            """)

        try:
            # Change can_return_rows_from_bulk_insert to be the same as when
            # has_trigger = True
            old_return_rows_flag = connection.features_class.can_return_rows_from_bulk_insert
            connection.features_class.can_return_rows_from_bulk_insert = False
            Author.objects.create(name='Foo')
        except django.db.utils.ProgrammingError as e:
            self.fail('Check for regression of issue #130. Insert with trigger failed with exception: %s' % e)
        finally:
            with connection.schema_editor() as cursor:
                cursor.execute("DROP TRIGGER TestTrigger")
            connection.features_class.can_return_rows_from_bulk_insert = old_return_rows_flag


class TestBinaryfieldGroupby(TestCase):
    def test_varbinary(self):
        with connection.cursor() as cursor:
            cursor.execute(f"SELECT binary FROM {BinaryData._meta.db_table} WHERE binary = %s GROUP BY binary", [bytes("ABC", 'utf-8')])


@skipUnlessDBFeature("supports_expression_defaults")
class DbDefaultBulkCreateRegressionTests(TransactionTestCase):
    """Regression tests for Django 6.0 db_default bulk insert alignment.

    Django 6.0 introduced DatabaseDefault sentinel values for fields with
    db_default. Our SQLInsertCompiler.as_sql() override must correctly:
      - Exclude a db_default column from the INSERT column list when *every*
        row uses the database default (letting the server supply the value).
      - Include the column when at least one row supplies an explicit value
        (mixing explicit values with DEFAULT keyword or prepared db_default).

    These tests create a raw table with a server-side DEFAULT SYSDATETIME()
    to exercise both paths end-to-end against SQL Server.
    """

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

        # This feature is Django 6.0+ only; skip cleanly on older versions.
        # Note: raise unittest.SkipTest, not cls.skipTest(), because
        # skipTest() is an instance method and cannot be called from
        # setUpClass (a classmethod).
        if VERSION < (6, 0):
            raise unittest.SkipTest(
                "db_default bulk insert alignment is Django 6.0+ specific"
            )

        # Unmanaged model so Django doesn't try to create/drop the table
        # via migrations — we handle that with raw SQL below.
        class DbDefaultBulkInsertModel(models.Model):
            name = models.CharField(max_length=100)
            created_at = models.DateTimeField(db_default=Now())

            class Meta:
                app_label = "testapp"
                db_table = "testapp_dbdefault_bulk_insert"
                managed = False

        cls.DbDefaultBulkInsertModel = DbDefaultBulkInsertModel

        with connection.cursor() as cursor:
            cursor.execute(
                """
                IF OBJECT_ID('testapp_dbdefault_bulk_insert', 'U') IS NOT NULL
                    DROP TABLE testapp_dbdefault_bulk_insert
                """
            )
            cursor.execute(
                """
                CREATE TABLE testapp_dbdefault_bulk_insert (
                    id INT IDENTITY(1,1) NOT NULL PRIMARY KEY,
                    name NVARCHAR(100) NOT NULL,
                    created_at DATETIME2 NOT NULL DEFAULT SYSDATETIME()
                )
                """
            )

    @classmethod
    def tearDownClass(cls):
        if VERSION < (6, 0):
            # Table was never created when the test was skipped.
            super().tearDownClass()
            return
        with connection.cursor() as cursor:
            cursor.execute(
                """
                IF OBJECT_ID('testapp_dbdefault_bulk_insert', 'U') IS NOT NULL
                    DROP TABLE testapp_dbdefault_bulk_insert
                """
            )
        super().tearDownClass()

    def test_db_default_field_excluded_and_included(self):
        """Verify created_at column presence in INSERT SQL for db_default.

        Case 1 — all rows use the database default:
          The column should be omitted from the INSERT column list so the
          server-side DEFAULT kicks in.  The quoted column name appears only
          in the OUTPUT INSERTED clause (if can_return_rows_from_bulk_insert)
          or not at all.

        Case 2 — at least one row supplies an explicit value:
          The column must appear in the INSERT column list (so the explicit
          value is written) *and* in OUTPUT INSERTED (if applicable).
        """
        model = self.DbDefaultBulkInsertModel
        created_at_quoted_name = connection.ops.quote_name("created_at")

        # --- Case 1: all rows rely on db_default for created_at ---
        with self.assertNumQueries(1) as ctx:
            model.objects.bulk_create([model(name="foo"), model(name="bar")])

        # created_at should NOT be in the INSERT column list.
        # It appears once only if OUTPUT INSERTED includes it.
        self.assertEqual(
            ctx[0]["sql"].count(created_at_quoted_name),
            1 if connection.features.can_return_rows_from_bulk_insert else 0,
        )

        # --- Case 2: one row overrides created_at with an explicit value ---
        with self.assertNumQueries(1) as ctx:
            model.objects.bulk_create(
                [
                    model(name="baz", created_at=timezone.now()),
                    model(name="qux"),
                ]
            )

        # created_at must be in the INSERT column list (1 occurrence)
        # plus OUTPUT INSERTED (1 more if can_return_rows_from_bulk_insert).
        self.assertEqual(
            ctx[0]["sql"].count(created_at_quoted_name),
            2 if connection.features.can_return_rows_from_bulk_insert else 1,
        )

    def test_single_insert_with_returning_fields_when_bulk_rows_unsupported(self):
        """Single-row create must still return all db_returning fields.

        Regression for a path where can_return_rows_from_bulk_insert=False
        caused SQLInsertCompiler to use SCOPE_IDENTITY() (1 column) while
        Django expected all returning fields, leading to IndexError.
        """
        model = self.DbDefaultBulkInsertModel
        old_return_rows_flag = connection.features_class.can_return_rows_from_bulk_insert
        connection.features_class.can_return_rows_from_bulk_insert = False
        try:
            with self.assertNumQueries(1) as ctx:
                obj = model.objects.create(name="single")

            self.assertIsNotNone(obj.pk)
            self.assertIsNotNone(obj.created_at)
            self.assertIn("OUTPUT INSERTED", ctx[0]["sql"])
            self.assertNotIn("SCOPE_IDENTITY", ctx[0]["sql"])
        finally:
            connection.features_class.can_return_rows_from_bulk_insert = old_return_rows_flag
